
------------------------------------------------------------------------------
-- An example template, showing how to define properties, services,
-- and tasks.
------------------------------------------------------------------------------

--
-- These require statements pull the specified shapes' functionality into the 
-- template. A shape can define properties, services, and tasks, just like a
-- template. If the template defines a property, service, or task with the
-- same name as one defined in a shape, then the definition in the shape will
-- be ignored.  Therefore, you must take care not to define characteristics
-- with duplicate names.
--

require "thingworx.shapes.filetransfer"
require "thingworx.shapes.propsubscribe"

--
-- This line is required in all user defined templates. 'template.example'
-- should be replaced with the name of template, for example: 'template.mydevice'.
--
module ("templates.modbusExample", thingworx.template.extend)

--
-- Properties are defined using Lua tables. The table format is as follows:
--     baseType: The ThingWorx base type for the property. Required.
--     dataChangeType: ALWYAS, VALUE, or NEVER. Each property in the properties table can be 
--                     monitored, and on change its value can be pushed to the server. A 
--                     dataChangeType of ALWAYS or VALUE should be used if the property is to 
--                     be pushed. Defaults to NEVER.
--     handler: The name of a handler script, http, https, or nil. The handler field defines the
--              script that will be used to get/set the property. It should contain the name of
--              a script executing within the script resource. If no handler is defined the 
--              property will be held in memory. Also see 'key' and 'value'.
--     key: A key that the handler can use to look up or set the property's value. In the case
--          of a script, this is a URL path. For http or https handlers, this field should contain
--          a URL, not including the protocol. Not required for handers that are nil.
--     setMethod: The HTTP method used by script handlers to get property. Defaults to GET.
--     getMethod: The HTTP method used by script handlers to set property. Defaults to SET.
--     value: The default value for the property. This will be updated as the property's value
--            changes during execution. Defaults to 0.
--     time: The last time, in milliseconds since the epoch, that the property was updated.
--           When Things are created from this template, this value will be automatically set to
--           the current time, unless a default is provided in the property definition.
--     quality: The quality of the property's value. A default should be provided if a default
--              value is defined. Otherwise it will default to GOOD for properties with no
--              handler, and UNKNOWN for properties with a handler.
-- 
properties.Int16InputRegExample = {key="input_register/1/30001?format=Int16", handler="modbus_handler", basetype="NUMBER"}
properties.Int16HoldRegExample = {key="holding_register/1/40001?format=Int16", handler="modbus_handler", basetype="NUMBER"}
properties.ScaledInt16RegExample = {key="holding_register/1/40001?format=Int16&scale=1.8&offset=32", handler="modbus_handler", basetype="NUMBER"}
properties.Uint16RegExample = {key="holding_register/1/40002?format=Uint16", handler="modbus_handler", basetype="NUMBER"}
properties.Int32RegExample = {key="holding_register/1/40003?format=Int32", handler="modbus_handler", basetype="NUMBER"}
properties.Uint32RegExample = {key="holding_register/1/40005?format=Uint32", handler="modbus_handler", basetype="NUMBER"}
properties.FloatRegExample = {key="holding_register/1/40007?format=Float", handler="modbus_handler", basetype="NUMBER"}
properties.StringRegExample = {key="holding_register/1/40009?format=String&length=6", handler="modbus_handler", basetype="STRING"}
properties.BitmaskRegExample = {key="holding_register/1/40015?format=MaskedBitmap16&offset=3&length=2", handler="modbus_handler", basetype="NUMBER"}
properties.CoilExample = {key="coil/1/1", handler="modbus_handler", basetype="BOOLEAN"}
properties.DiscreteExample = {key="discrete_input/1/10003", handler="modbus_handler", basetype="BOOLEAN"}

--
-- Services are defined as Lua functions. These can be executed remotely from the server and must 
-- provide a valid response, via their return statement. The signature for the functions must be:
--   me: A table the refers to the Thing.
--   headers: A table of HTTP headers.
--   query: The query paramters from the HTTP request.
--   data: A Lua table containing the parameters to the service call.
-- 
-- A service must return the following values (in this order):
--   * A HTTP return code (200 for success)
--   * A table of HTTP response headers. This should contain a valid Content-Type header, typically
--     with a value of application/json. A default table can be easily generated by calling
--     tw_utils.RESP_HEADERS().
--   * The response date, in the form of a JSON string. This can be created from a Lua table using 
--     json.encode, or tw_utils.encodeData().
-- 

services.getAllProperties = function(me, headers, query, data) 
  log.debug(me.name, "Called getProperties")
  -- Use a JSON array for property formats for the request
  local postdata = {content='{"slave":1,"address":"40001","count":16, "formats":[["Int16", 1.8, 32], "Uint16", "Int32", "Uint32", "Float",["String",6], ["MaskedBitmap16", 3, 2]]}'}
  return tw_script.executeCallback("modbus_handler", "POST", "read_multiple_registers", headers, query, postdata)
end


--
-- Tasks are Lua functions that are executed periodically by the ThingWorx Lua framework. They
-- can be used to execute background tasks, monitor resources, and fire events.
--[[ 
tasks.errorRule = function(me)
  -- Do task
end
--]]
