-- Modbus interface library
--
-- Developed by ThingWorx
-- (c)2012 ThingWorx
--
------------------------------------------------------------------
------------------------------------------------------------------
--[[
Each modbus read request should have a "formats" table associated
with it.  The formats table is a numerically indexed table of tables.
Each entry should have a format type identifer plus any extra data
required to do the formatting.  Valid types and extra data are:

Signed 16 bit number: Int16, scale(optional), offset(optional)
Unsigned 16 bit number: Uint16, scale(optional), offset(optional)
Signed 32 bit number: Int32, scale(optional), offset(optional)
Unsigned 32 bit number: Uint32, scale(optional), offset(optional)
Single precison Float: Float, scale(optional), offset(optional)
Double precison Float: Double
String:	String, max_length
16 bit wide Masked Bitmap: MaskedBitmap16, mask_offset, mask_length
32 bit wide Masked Bitmap: MaskedBitmap32, mask_offset, mask_length
Linux Timestamp: Timestamp

Note that all numeric values are retuned as Lua numbers, so with scaling &
offset and Integer may not actually result in an Integer value.
Timestamps consist of 6 contiguous 16 bit registers that contain, in order,
Month, Day, Year, Hour, Minute, Milliseconds
--]]
------------------------------------------------------------------
------------------------------------------------------------------
local base = _G
local type, pairs, error, ipairs, string, unpack, table, math, tonumber, tostring, package, print =
	   type, pairs, error, ipairs, string, unpack, table, math, tonumber, tostring, package, print
local p_data, tw_logger, tw_utils, tw_mutex, tw_convert = p_data, tw_logger, tw_utils, tw_mutex, tw_convert

module("modbus")

mbErrCodes = {
	[0x00] = "00 UNKNOWN",
	[0x01] = "01 ILLEGAL FUNCTlON",
	[0x02] = "02 ILLEGAL DATA ADDRESS",
	[0x03] = "03 ILLEGAL DATA VALUE",
	[0x04] = "04 ILLEGAL RESPONSE LENGTH",
	[0x05] = "05 ACKNOWLEDGE",
	[0x06] = "06 SLAVE DEVlCE BUSY",
	[0x07] = "07 NEGATlVE ACKNOWLEDGE",
	[0x08] = "08 MEMORY PARlTY ERROR",
	[0x09] = "09 UNKNOWN",
	[0x0A] = "0A GATEWAY PATH UNAVAILABLE",
	[0x0B] = "0B GATEWAY TARGET DEVICE FAILED TO RESPOND",
}

mbFunctionCodes = {
	READ_COIL_STATUS = 0x01,
	READ_DISCRETE_STATUS = 0x02,
	READ_HOLDING_REGISTERS = 0x03,
	READ_INPUT_REGISTERS = 0x04,
	WRITE_SINGLE_COIL = 0x05,
	WRITE_SINGLE_REGISTER = 0x06,
	WRITE_MULTIPLE_COILS = 0x0F,
	WRITE_MULTIPLE_REGISTERS = 0x10,
}

mbFormatFunctions = {
	Int16 = tw_convert.toInt16,
	Uint16 = tw_convert.toUint16,
	Int32 = tw_convert.toInt32,
	Uint32 = tw_convert.toUint32,
	Float = tw_convert.toFloat,
	Double = tw_convert.toDouble,
	String = tw_convert.toString,
	MaskedBitmap16 = tw_convert.toMaskedBitmap16,
	MaskedBitmap32 = tw_convert.toMaskedBitmap32,
	Timestamp = tw_convert.toTimestamp,
}

modbus = {}
modbus.__index = modbus
log = base.require("log")
------------------------------------------------------------------
-- new
-- Creates a new modbus object.
-- Parameters:
-- Returns:
-- 		The newly created object
------------------------------------------------------------------
function modbus:new (config)
	-- config is a table that determines whether this is RTU or TCP
	local e -- error string
	if type(config) ~= "table" then
		log.error("modbus:new", string.format("Input parameter must be a table"))
		return nil
	end
	local me = {}   -- create object if user does not provide one
	base.setmetatable(me, modbus)
	me.response_timeout = config.response_timeout or 5000
	me.rcv_state = "IDLE"
	me.state = "IDLE" -- IDLE, TRANSMITTING, WAITING_FOR_RESPONSE
	me.reg_adjust = config.reg_adjust or 1
	-- Now determine if this is RTU or TCP
	if config.host then
		-- This is TCP
		socket = base.require("socket")
		-- If we were already open we should close and reopen
		if self.p then p:close() end
		me.tcp, e = socket.tcp()
		if not me.tcp then
			log.error("modbus:new","Can't create a socket.  Error: %s", err or "unknown")
			return nil
		end
		me.host = config.host;
		me.port = config.port or 502
		me.read_timeout = config.read_timeout or 1
		me.tcp:settimeout(me.read_timeout)
		me.p = {}
		me.p.close = me.tcp.close
		-- Set up our functions so whether we are tcp or rtu is abstracted the rest of the way
		me.p.write = function(empty, frame)
			if not me.p.connected then
				log.debug("modbus:p.write", "Attempting to connect via ModbusTCP...")
			    me.p.connected, e = me.tcp:connect(me.host, me.port)
				if not me.p.connected then
					if e and e == "already connected" then
						me.p.connected = true
					else
						log.error("modbus:tcp.write", "Can't open socket.  Error: %s", e or "unknown")
						return nil, "Error opening socket"
					end	
				end
			end
			local a,b = me.tcp:send(frame)
			if b and b == "closed" then 
				log.debug("modbus:p.write", "Socket was closed, tearing it down and creating a new one.")
				me.tcp, e = socket.tcp()
				if not me.tcp then
					log.error("modbus:p.write","Can't create a socket.  Error: %s", err or "unknown")
					return nil
				end			
				me.p.connected = nil 
			end
			return b,a
		end
		me.p.read = function(self,cnt)
			local data_read, e = me.tcp:receive(cnt)
			if data_read then return nil, data_read, data_read:len() 
			else return e, -1, nil end
		end
	elseif config.serial_port then
		-- This is RTU so load the RS232 library
		me.rs232 = base.require("luars232")
		-- Set our defaults
		me.read_timeout = config.read_timeout or 50
		local default_port = "ttyS0"
		if tw_utils.host_os() == "win32" then default_port = "com1" end
		me.serial_port = config.serial_port or default_port
		if not config.baud_rate then me.baud_rate = me.rs232.RS232_BAUD_9600
		elseif config.baud_rate == "1200" then me.baud_rate = me.rs232.RS232_BAUD_1200
		elseif config.baud_rate == "2400" then me.baud_rate = rs232.RS232_BAUD_2400
		elseif config.baud_rate == "4800" then me.baud_rate = me.rs232.RS232_BAUD_4800
		elseif config.baud_rate == "9600" then me.baud_rate = me.rs232.RS232_BAUD_9600
		elseif config.baud_rate == "19200" then me.baud_rate = me.rs232.RS232_BAUD_19200
		elseif config.baud_rate == "38400" then me.baud_rate = me.rs232.RS232_BAUD_38400
		elseif config.baud_rate == "57600" then me.baud_rate = me.rs232.RS232_BAUD_57600
		elseif config.baud_rate == "115200" then me.baud_rate = me.rs232.RS232_BAUD_115200
		else me.baud_rate = me.rs232.RS232_BAUD_9600 end

		-- If we were already open we should close and reopen
		if me.p then me.p.close() end
		local os = tw_utils.host_os()
		if os == "linux" then
			me.serial_port = "/dev/" .. me.serial_port
		elseif os == "win32" then
			me.serial_port = "\\\\.\\" .. me.serial_port
		end
		e, p = me.rs232.open(me.serial_port)
		if e ~= me.rs232.RS232_ERR_NOERROR then
			log.error("modbus:new", string.format("can't open serial port '%s', error: '%s'",
					me.serial_port, me.rs232.error_tostring(e)))
			return nil
		end
		-- Set our baud rate & other port settings
		if me.baud_rate then p:set_baud_rate(me.baud_rate) end
		if config.data_bits then p:set_data_bits(config.data_bits) end
		if config.parity then p:set_parity(config.parity) end
		if config.stop_bits then p:set_stop_bits(config.stop_bits) end
		if config.flow_control then p:set_flow_control(config.flow_control) end
		me.p = p
		if not me.rs232 then
			log.error("modbus:new", string.format("can't create serial port '%s'", me.serial_port))
			return nil
		end
	else
		-- This is just wrong
		log.error("modbus:new", string.format("Configuration must include either a host or a serial_port"))
		return nil
	end
	return me
end

------------------------------------------------------------------
-- modbus:transmitFrame
-- Create and send an modbus frame.
-- Parameters:
--      slave - destination slave
--      func - the function to request
--      register - the base register
--      count - the number of registers or coils
--      data - data to send
-- Returns:
-- 		response frame or nil plus error string upon failure
------------------------------------------------------------------
function modbus:transmitFrame(slave, func, register, count, data)
	if type(slave) ~= "number" or slave < 0 or slave > 255 or
	  (type(register) ~= "number" or register < 0 or register > 65535) or
	  count > 125 or --This is not entirely correct for coils but keeps things simple for now
	  (data and type(data) ~= "string") then
		local err = "Bad paramter(s).  Must have a valid Slave ID, function code, register & count.  Data, if present must be a string. slave:"..slave
		log.error("modbus:transmitFrame", err)
		return nil, err
	end
	
	log.debug ("modbus:transmitFrame","slave = %d, function = %d, register = %d, count = %d", slave, func, register, count)
	-- Check our state
	if self.state ~= "IDLE" then
		return nil, "BUSY"
	else
		self.state = "TRANSMITTING"
	end
	local temp = {}
	local length = 5
	temp[1] = slave
	temp[2] = func
	temp[3] = math.floor(register / 256)
	temp[4] = register % 256
	-- If writing the next few bytes mean different things
	local data_start = 0x05
	if func == 0x10 then
		temp[5] = math.floor(count / 256)
		temp[6]	= count	% 256
		temp[7]	= count * 2
		data_start = 0x07
	elseif func == 0x06 then
		-- SIngle register, no count
		data_start = 0x04
	elseif func == 0x05 then
		-- Single coil, no count
		data_start = 0x04
	else
	    -- Not a write, just add count
		temp[5] = math.floor(count / 256)
		temp[6]	= count	% 256
		data_start = 0x06
	end
	-- Set our data, if any
	if type(data) == "string" then
		for i = 1, data:len() do
			temp[i + data_start] = data:byte(i)
			length = length + 1
		end
	end
	
	if func == 0x05 then
		-- Single coil, has leading zeros at end
		table.insert(temp, 0x00)
		length = length + 1
	end
	
	if self.rs232 then
		-- add our CRC if we are using RTU
		local crc = tw_utils.calculateCRC16(string.char(unpack(temp)))
		table.insert(temp, math.floor(crc % 256))
		table.insert(temp, math.floor(crc/256))
------		table.insert(temp, math.floor(crc % 256))
	elseif self.tcp then
		-- prepend our tcp stuff
		local len = #temp
		for i = 1, 5 do
			table.insert(temp,1,0)
		end
		table.insert(temp,6,len)
	end
	local frame = string.char(unpack(temp))
	-- ###### FOR DEBUGGING - CAN BE REMOVED ######
	log.data_bytes("modbus:transmitFrame", frame)
	-- ############################################
	-- send the frame
	if (self.p) then
		local e, len_written = self.p:write(frame)
		self.state = "WAITING_FOR_RESPONSE"
		if len_written == frame:len() then
			log.debug("modbus:transmitFrame","Waiting for Response")
			return frame:len()
		end
		log.error("modbus:transmitFrame", "Error sending frame. Err: %s", e or "unknown")
		return nil, e or "unknown"
	else
		log.error("modbus:transmitFrame", "No transport available")
		self.state = "IDLE"
		return nil, "No transport available"
	end
end
------------------------------------------------------------------
------------------------------------------------------------------
-- modbus:rcvFrame(slaveId, code)
-- Receive a response frame.
-- Parameters:
--      slaveId - use this to quickly terminate an out of sync frame
--      code - expected code. use this to quickly terminate an out of sync frame
-- Returns:
--      table newframe (or nil if no frame is available or an error occured)
--          newframe contains the following
--      	slave - the slave that sent the response
--			code - the function code
--			data - the data in the response
------------------------------------------------------------------
function modbus:rcvFrame(slaveId, code)
	-- Read one byte at a time until we get all the data or a timeout
	-- Receive states IDLE, CODE, COUNT, DATA, CRC_HI, CRC_LO
	if type(slaveId) ~= "number" or slaveId < 0 or slaveId > 255 then
		log.warn("modbus:rcvFrame", "Invalid Slave Id")
		self.state = "IDLE"
		return nil
	end
	if not self.p then
		log.warn("modbus:rcvFrame", "Transport is not initialized'")
		self.state = "IDLE"
		return nil
	end
	local e
	local resp = {}
	local frame = ""
	local data_read = ""
	local size = 0
	local count = 0
	local crc = ""
	---
	-- If we are using Modbus TCP we need to read and dispose of the Modbus TCP header
	if self.tcp then
		log.debug("modbus:rcvFrame", "Reading Modbus TCP Header")
		e, data_read, size = self.p:read(6, self.read_timeout)
		if size == 0 or data_read == nil then
			log.error("modbus:rcvFrame", "Error reading from transport. Read %d bytes, e = %s", size or 0, e or "nil")
			self.state = "IDLE"
			return nil, e
		end
	end
	-- Check to see if real Modbus data is available.
	e, data_read, size  = self.p:read(1, self.read_timeout)
	if size == 0 or data_read == nil or data_read == -1 then
--		log.trace("modbus:rcvFrame", "Read %d bytes, e = %s", size or -1, e or "nil")
		self.state = "IDLE"
		return nil, e
	end
	self.rcv_state = "SLAVE"
	while data_read do
		if self.rcv_state == "SLAVE" then
			log.debug("modbus:rcvFrame", "Got Slave Id = %d", data_read:byte(1))
			-- Check to see if we got the right slave
			if data_read:byte(1) ~= slaveId then
				log.error("modbus:rcvFrame", "Received %d for Slave Id instead of expected %d",
					data_read:byte(1), slaveId )
				self.rcv_state = "IDLE"
				-- Drain any remaining data
				size = 1
				while size == 1 do
				    data_read = nil
					e, data_read, size  = self.p:read(1, self.read_timeout)
				end
				self.state = "IDLE"
				return nil, "Invalid slave"
			end
			resp.slave = slaveId
			frame = frame .. data_read
			self.rcv_state = "CODE"
		elseif self.rcv_state == "CODE" then
			log.debug("modbus:rcvFrame", "Got Code = %d", data_read:byte(1))
			resp.code = data_read:byte(1)
			frame = frame .. data_read
			if data_read:byte(1) ~= code then
				log.error("modbus:rcvFrame", "Received %d for Function code instead of expected %d",
					data_read:byte(1), code )
				-- If it is an error code, print the error
				if data_read:byte(1) == code + 0x80 then
					e, data_read, size  = self.p:read(1, self.read_timeout)
					data_read = data_read or "?"
					log.error("modbus:rcvFrame", "Modbus Error: %s", mbErrCodes[data_read:byte(1)] or "unknown")
					e = mbErrCodes[data_read:byte(1)] or "unknown"
				end
				self.rcv_state = "IDLE"
				-- Drain any remaining data
				while data_read do
				    local tmp
				    size, data_read, tmp = self.p:read(1, self.read_timeout)
				    
					-- RL Updated 9/24/2014 to avoid getting stuck in this loop
				    if size == "timeout" and data_read == -1 then
						data_read = ""
						log.error("modbus:rcvFrame", "Received a timeout, exiting the loop that drains data_read");
					break
				    end

				end
				self.state = "IDLE"
				return nil, e
			end
			-- Rest of the state machine depends on what the code was
			if code == 0x05 or code == 0x06 or code == 0x0F or code == 0x10 then
			    -- This was a write.  If we are here the write succeeded
				-- with a four byte response that has differnet meanings
				e, data_read, size  = self.p:read(4, self.read_timeout)
				-- TODO - in the future we may want ot validate the response
				frame = frame .. data_read
				-- ###### FOR DEBUGGING - CAN BE REMOVED ######
				log.data_bytes("modbus:responseFrame", frame)
				-- ############################################
				resp.data = data_read
				-- If we are using Modbus TCP there is no CRC
				if self.tcp then
					--self.tcp:close()
					self.rcv_state = "IDLE"
					self.state = "IDLE"
					return resp
				else
					self.rcv_state = "CRC_LO"
				end
			else
				self.rcv_state = "COUNT"
			end
			--frame = frame .. data_read
		elseif self.rcv_state == "COUNT" then
			count = data_read:byte(1)
			frame = frame .. data_read
			self.rcv_state = "DATA"
		    e, data_read, size  = self.p:read(count, self.read_timeout)
			if size < count then
				log.error("modbus:rcvFrame", "Error reading data. Error: %s", e or "unknown")
				self.rcv_state = "IDLE"
				return nil, e or "unknown"
			end
			frame = frame .. data_read
			-- ###### FOR DEBUGGING - CAN BE REMOVED ######
			log.data_bytes("modbus:responseFrame", frame)
			-- ############################################
			resp.data = data_read
			-- If we are using Modbus TCP there is no CRC
			if self.tcp then
				--self.tcp:close()
				self.rcv_state = "IDLE"
				self.state = "IDLE"
				return resp
			else
				self.rcv_state = "CRC_LO"
			end
		elseif self.rcv_state == "CRC_LO" then
			crc = data_read:byte(1)
			self.rcv_state = "CRC_HI"
		elseif self.rcv_state == "CRC_HI" then
			crc = crc + data_read:byte(1)  * 256
			if tw_utils.calculateCRC16(frame) ~= crc then
				resp = nil
				log.error("modbus:rcvFrame", "CRC check failed.  Actual: %x, Calculated: %x", crc, tw_utils.calculateCRC16(frame) or 0 )
				self.rcv_state = "IDLE"
				self.state = "IDLE"
				return nil, "CRC check failed"
			end
			-- Drain any remaining data
			size = 1
			while size == 1 do
				data_read = nil
				e, data_read, size  = self.p:read(1, self.read_timeout)
			end
			self.rcv_state = "IDLE"
			self.state = "IDLE"
			return resp
		else
			log.error("modbus:rcvFrame", "Invalid receive state %s", self.rcv_state or "nil" )
			self.rcv_state = "IDLE"
			-- Drain any remaining data
			while data_read do
				local tmp
				data_read = nil
				size, data_read, tmp = self.p:read(1, self.read_timeout)
			end
			self.state = "IDLE"
			return nil, "Invalid receive state"
		end
		e, data_read, size  = self.p:read(1, self.read_timeout)
		if size == 0 or data_read == nil then
			log.error("modbus:rcvFrame", "Error reading from transport. Read %d bytes, e = %s", size or 0, e or "nil")
			self.state = "IDLE"
			return nil, e
		end
	end
end
------------------------------------------------------------------
------------------------------------------------------------------
-- modbus:executeFunction
-- Create and send a Modbus frame and wait for the response.
-- Parameters:
--      slave - destination slave
--      func - the function to request
--      register - the base register
--      count - the number of registers or coils
--      data - data to send
--		formats - a table with the format specifiers of the requested registers
-- Returns:
-- 		Response data on success or nil plus error string upon failure
--		If function is a write, a 0 is returned on success
------------------------------------------------------------------
function modbus:executeFunction(slave, func, register, count, data, formats)
	-- Check for a valid command
	if type(slave) ~= "number" or slave < 0 or slave > 255 or
	  not mbFunctionCodes[func] or
	  (type(register) ~= "number" or register > 65535) or
	  count > 125 or --This is not entirely correct for coils but keeps things simple for now
	  (data and type(data) ~= "string") then
		local err = "Bad paramter(s).  Must have a valid Slave ID, function code, register & count."
		err = err .. "Data, if present must be a string"
		log.error("modbus:executeFunction", err)
		return nil, err
	end
	-- Make sure we aren't already transmitting
	if self.state ~= "IDLE" then
		log.error("modbus:executeFunction", "Another command is already in progress")
		return nil, "Another command is already in progress"
	end
	-- Adjust our registers
	func = mbFunctionCodes[func]
	if func == 0x02 and register > 10000 then register = register - 10000
	elseif func == 0x04 and register > 30000 then register = register - 30000
	elseif (func == 0x03 or func == 0x06 or func == 0x10 or func == 0x16) and register > 40000 then register = register - 40000
	----else register = register - 1  -- Still need to offset the 1 for register counting
	end
	register = register - self.reg_adjust
	-- Transmit the command frame
	local result, e = self:transmitFrame(slave, func, register, count, data)
	if not result then
		log.error("modbus:executeFunction", "Error transmitting frame.  Error: %s", e or "unknown")
		self.state = "IDLE"
		return nil, e
	end
	-- Get our result
	local starttime = tw_utils.currentTime()
	-- Listen for a result frame
	e, result = nil
	log.debug("modbus:executeFunction", "Waiting for response for Frame")
	while (not (result or e) and tw_utils.currentTime() < starttime + self.response_timeout) do
		result, e = self:rcvFrame(slave, func)
		if type(result) == "table" then
			self.state = "IDLE"
			if result.slave == slave and result.code == func then
				log.trace("modbus:executeFunction", "Received frame, processing response")
				return self:processResponse(func, result.data, formats)
			else
				log.warn("modbus:executeFunction", "Invalid Slave or Function Code in result")
				return nil, "Invalid Slave or Function Code in result"
			end
		elseif not result then
			if e and (e == "timeout" or e == self.rs232.RS232_ERR_NOERROR or e == self.rs232.RS232_ERR_TIMEOUT) then
			-- if e and (e == "timeout") then
				e = nil
			else	
				self.state = "IDLE"
				log.warn("modbus:executeFunction", "Invalid result type or error.  Res type: %s Err: %s", type(res), e or "nil")
				return nil, "Invalid result type"
			end
		end
	end
	self.state = "IDLE"
	if e then
		log.error("modbus:executeFunction", "Error receiving result.  Error: %s", e)
		return nil, e
	else
		log.error("modbus:executeFunction", "Error: Timed out waiting for result.")
		return nil, "Timeout"
	end
end
------------------------------------------------------------------
------------------------------------------------------------------
-- processResponse
-- Processes a received frame.
-- Parameters:
--		code - the original funciton code
--      pkt - a string containing the raw bytes in a response
--		formats - a numerically indexed table representing how to format the response
-- Returns:
--      values - a numerically indexed table with the transformed values. Any valid
--				values that are calulated before the erro will be returned in the table
--		err - indication that an erro occured and this may be a partial table
------------------------------------------------------------------
function modbus:processResponse(code, pkt, formats)
	-- Do some basic validation on the input
	-- ###### FOR DEBUGGING - CAN BE REMOVED ######
	log.data_bytes("processResponse", pkt)
	-- ############################################
	-- If this was a write and we are here, the write was successful
	-- so just return a 0
	if code==0x05 or code==0x0F or code==0x06 or code==0x10 or code==0x16 or code==0x15 then
		log.debug("modbus:processResponse", "Function was a write, returning 0 for success")
		return {[1] = 0}
	end
	-- If we don't have a formats table, just return the raw bytes
	if type(formats) ~= "table" then
		log.debug("modbus:processResponse", "No formats specified, returning raw bytes")
		return { [1] = pkt }
	end
	-- Now do the formatting as specified in the formats table.
	-- Each entry will have a format type plus a table with any additoinal
	-- formatting information described at the beginning of this document
	result = {}
	for _,v in ipairs(formats) do
		log.trace("modbus:processResponse", "Remaining bytes: %d", pkt:len())
		-- ###### FOR DEBUGGING - CAN BE REMOVED ######
		log.data_bytes("processResponse", pkt)
		-- ############################################
		if type(v) ~= "table" then
			if mbFormatFunctions[v] and pkt:len() > 0 then
				local res, bytesUsed = mbFormatFunctions[v](pkt)
				if res then
					table.insert(result, res)
					pkt = pkt:sub(bytesUsed + 1)
				else
					log.error("modbus:processResponse", "Conversion function failed.  Error: %s",
						bytesUsed or "nil")
					return result, bytesUsed
				end
			end
		elseif mbFormatFunctions[v[1]] and pkt:len() > 0 then
			local temp =  {}
			if v[2] then temp[1] = v[2] end
			if v[3] then temp[2] = v[3] end
			if v[4] then temp[3] = v[4] end
			local res, bytesUsed = mbFormatFunctions[v[1]](pkt, unpack(temp))
			if res then
				table.insert(result, res)
				pkt = pkt:sub(bytesUsed + 1)
			else
				log.error("modbus:processResponse", "Conversion function failed.  Error: %s",
					bytesUsed or "nil")
				return result, bytesUsed
			end
		else
			log.error("modbus:processResponse", "Invalid conversion function")
			return result, "Invalid conversion function"
		end
	end
	return result, nil
end
------------------------------------------------------------------
function modbus:close()
	if self.p then self.p:close() self.p = nil end
end

function modbus:set_read_timeout(x)
	self.read_timeout = x
end

return {
    init = function(t)
		return modbus:new(t)
    end
}
