--
-- win-registry
-- Provides windows registry access
-- Requires LuaCOM
--
-- NOT A STANDALONE SCRIPT, THIS SCRIPT
-- MUST BE REQUIRED BY SYSTME_WMI.LUA
--
-- Developed by Palantiri Systems
-- (c)2011 Palantiri Systems, Inc.
--
-- Configuration Items
--   None
-- Usage
--   REST Based interface
--		GET /registry/<tree>/<key>/[<valueName>]
--			Retrieves the full depth of a key or a single value
--			if a value name is specified
--		PUT /registry/<tree>/<key>
--			Creates the key.  Creates all required parent keys as well.
--		DELETE /registry/<tree>/<key>/[<valueName>]
--			Deletes the key.  If this is a key then all subkeys 
--			must have been previously deleted.
--		POST /registry/<tree>/<key>/<valueName>
--			Updates (or creates) the value in the specified key.  
--			Required POST parameters are:
--	 			value = <value>  
--					(REG_BINARY values are space separated hex digits)
--			Optional POST parameters are:
--				vtype = <REG_SZ|REG_DWORD|REG_EXPAND_SZ|REG_BINARY>
--					If not specified, the current type is used.  If
--                  this is a new key, then the default is REG_SZ
------------------------------------------------------------------
------------------------------------------------------------------
-- A function to load the ps_* libraries.  THIS IS TO BE USED WHEN
-- RUNNING IN A STANDALONE DEBUGGER ONLY.  DO NOT CALL THIS FUNCTION
-- WHEN RUNNING IN THE PALANTIRI LUA SCRIPT RESOURCE.
-- Make sure to set you path to the proper value
function init_ps_library()
-- function to load a DLL
local path = "C:\\1Palantiri\\Product\\trunk\\extensions\\luaScripts\\debug_ps_library\\ps_library.dll"
local f = assert(package.loadlib(path, "luaopen_ps_library"))
f()
package.path = 'C:\\1Palantiri\\Product\\trunk\\extensions\\luaScripts\\lua_libs\\?.lua;' .. package.path
package.cpath = 'C:\\1Palantiri\\Product\\trunk\\extensions\\luaScripts\\lua_clibs\\win32\\?.dll;' .. package.path
ps_logger.force("lua", "loaded ps_library")
ps_debugging = true
end
--------------------------------------------------------------------
trees = {
		HKEY_CLASSES_ROOT = 0x80000000,
		HKEY_CURRENT_USER = 0x80000001,
		HKEY_LOCAL_MACHINE = 0x80000002,
		HKEY_USERS = 0x80000003,
        HKEY_CURRENT_CONFIG = 0x80000005,
		HKCR = 0x80000000,
		HKCU = 0x80000001,
		HKLM = 0x80000002,
}

reg_types = {
		[1]  = "REG_SZ",
		[2]  = "REG_EXPAND_SZ",
		[3]  = "REG_BINARY",
		[4]  = "REG_DWORD",
		[7]  = "REG_MULTI_SZ",
		[11] = "REG_QWORD",
}

reg_type_string = {
		REG_SZ = 1,
		REG_EXPAND_SZ = 2,
		REG_BINARY = 3,
		REG_DWORD = 4,
		REG_MULTI_SZ = 7,
		REG_QWORD = 11,
}

------------------------------------------------------------------
-- processRestRequest
-- REST interface that gets, sets or deletes the specified key
-- Parameters:
--    method - a string containing the method requested (GET, POST, DELETE)
--    path - an array of strings containing the HTTP path specified in the request "/my/path/to/data"
--    headers - a table of HTTP header attribute/value pairs in the request, can be nil
--    query - a table of attribute/value pairs in any query string ("?myattribute=myvalue"), can be nil
--    postdata - a table of attribute/value pairs in post data ("myattribute=myvalue"), can be nil
-- Returns: ResponseCode, Headers, Value
------------------------------------------------------------------
function processRestRequest(method, path, headers, query, postdata)
	log.debug("processRestRequest", "Processing request %s", table.concat(path, '/'))
	-- Make sure we have a valid tree
	-- get rid of the first path entry ('registry')
	table.remove(path,1)
	if #path < 2 or not trees[path[1]] then
		return 404, "Invalid key"
	end
	local key = path[2]
	local valueName = path[#path]
	for i = 3, #path do
		key = string.format("%s\\%s", key, path[i])
	end
	-- COmmon request info
	local req = { tree = path[1], key = key }
  local contentType = "application/json"
	if (method == "GET") then
		-- Set our content type
		if headers and headers["content-type"] then contentType = headers["content-type"] end
		if query and query["format"] then
			contentType = query["format"]
			headers["content-type"] = query["format"]
		end
		req.cmd = "read"
	elseif (method == "POST") then
		req.cmd = "update"
		if postdata and postdata.value then
			req.cmd = "update"
			req.value = postdata.value
			req.vtype = postdata.vtype
		else
			return 400, "Must specify a value"
		end
	elseif (method == "PUT") then req.cmd = "create"
	elseif (method == "DELETE") then req.cmd = "delete"
	else
		log.warn("processRestRequest", method .. " method is not supported")
		return 405, "Only GET, POST, PUT, DELETE methods supported"
	end
	-- Make the request
	local code, result = makeRequest(req)
	if code ~= 200 then
		log.error("processRestRequest", "Error retrieving value for %s\\%s", path[1], key)
		return code, tostring(result) or "unknown error"
	end
	log.trace("processRestRequest", "Value for %s\\%s = %s", path[1], key, table.toString(result))
	return 200, headers, table.toContentTypeString(result, contentType, query)
end
------------------------------------------------------------------
------------------------------------------------------------------
-- makeRequest
-- Queues the request up for processing and waits for the result
-- Parameters:
--    req - table containg cmd, tree, key, valueName, newValue
-- Returns: the resulting table or nil if error
------------------------------------------------------------------
function makeRequest(req)
	if not req then
		log.error("makeRequest", "nil request received")
		return nil
	end
	if not req.cmd or not req.tree or not req.key then
		log.error("makeRequest", "nil cmd, tree or key received")
		return nil
	end
	-- We need to pass this to the main loop for execution.  COM requires
	-- that each thread call CoInitialize, but this is only called when
	-- we do the "require Luacom".  So the callback threads have not
	-- called CoInitialize and there is no easy/efficient way to reload
	-- the library.  So passing the request to the main loop is the best
	-- way.
	local truncated
	-- Create a timeout so we don't wait for a result forever
	local now = ps_utils.currentTime()
	ps_mutex.lock()
	req_id = req_id + 1
	log.debug("makeRequest", "Called.  ReqID: %d.  Command: %s.  Key: %s\\%s\\%s", req_id, req.cmd, req.tree, req.key, (req.valueName or ""))
	registry_req = req
	registry_resp = nil
	-- Wait for a result or timeout
	while registry_resp == nil and ps_utils.currentTime() < now + p_data.request_timeout do
		ps_utils.psleep(10)
	end
	local result, code
	if registry_resp then
		if type(registry_resp) ~= "table" then
			log.error("makeRequest", "ReqID: " .. req_id .. " Got error result from resp queue")
			code = 500
			result = tostring(registry_resp)
		else
			code = 200
			result = table.fullCopy(registry_resp)
		end
	else
		-- We timed out
		log.error("makeRequest", "ReqID: " .. req_id .. " timed out")
		code = 408
		result = "Query timed out"
	end
	ps_mutex.unlock()
	return code, result
end
------------------------------------------------------------------
------------------------------------------------------------------
-- keyToTable
-- Iterates through a registry key an maps subkeys and values
-- into table entries. The resulting table contains two types of
-- subtables, subkeys and values.
-- Parameters:
--    svc - the LuaCom object to use
--    tree - the registry key (hive) to check
--    key - the key to operate on
-- Returns: the resulting table or nil plus error message if error
------------------------------------------------------------------
function keyToTable(svc, tree, key)
	if not key or not svc or not tree then
		log.error("keyToTable", "nil request received")
		return nil
	end
	local keytable = {}
	local result
	-- Get any values that may exist
	local ValueNames, ValueTypes
	log.trace("keyToTable", "Reading key %s\\%s\\", tree, key)
	luacom.config.last_error = nil
	result, _, _, ValueNames, ValueTypes = svc:EnumValues(trees[tree], key, ValueNames, ValueTypes)
	if result then
		if result == 0 then
			-- Add the stadard default, which may or may not be overwritten
			keytable["(Default)"] = { vtype = "REG_SZ", value = "" }
			if type(ValueNames) == "table" and type(ValueTypes) == "table" then
				for i = 1, #ValueNames do
					local valtype = reg_types[ValueTypes[i]] or "unknown"
					local current_val
					-- We got the type, now get the value
					if valtype == "REG_DWORD" then
						result, _, _, _, current_val = svc:GetDWORDValue(trees[tree], key, ValueNames[i], _)
					elseif valtype == "REG_SZ" then
						result, _, _, _, current_val = svc:GetStringValue(trees[tree], key, ValueNames[i], _)
					elseif valtype == "REG_EXPAND_SZ" then
						result, _, _, _, current_val = svc:GetExpandedStringValue(trees[tree], key, ValueNames[i], _)
					elseif valtype == "REG_BINARY" then
						result, _, _, _, current_val = svc:GetBinaryValue(trees[tree], key, ValueNames[i], _)
						-- Convert the values to hex
						for i = 1, #current_val do current_val[i] = string.format("%02x", current_val[i]) end
					elseif valtype == "REG_MULTI_SZ" then
						result, _, _, _, current_val = svc:GetMultiStringValue(trees[tree], key, ValueNames[i], _)
					elseif valtype == "REG_QWORD" then
						-- Some clients may not support QWORD
						if svc.GetQWORDValue then
							result, _, _, _, current_val = svc:GetQWORDValue(trees[tree], key, ValueNames[i], _)
						else
							log.warn("keyToTable", "QWORD value of %s\\%s\\%s.  Type not supported", tree, key, ValueNames[i])
							result = 0
							current_val = "unable to read QWORD values"
						end
					end
					if result and result == 0 then
						keytable[ValueNames[i]] = { vtype = valtype, value = current_val }
					else
						log.error("keyToTable", "Error getting value of %s\\%s\\%s.  Error: %s",
							tree, key, ValueNames[i],luacom.config.last_error or tostring(result))
						keytable[ValueNames[i]] = { vtype = valtype, value = "error reading value" }
					end
				end
			end
		elseif result == 2 then
			-- We may have been passed a value name.  Try to retrieve it by
			-- chopping off the last part of the key variable to get all the
			-- key values and then select the one we are looking for
			local _, _, name = string.find(key, "([^\\]*)$")
			key = string.gsub(key, "[^\\]*$", "")
			local temp, msg = keyToTable(svc, tree, key)
			log.debug("keyToTable", "Trying key %s with value name = %s", key, name)
			if temp and temp[name] then
                if type(temp[name]["value"]) == 'table' then
                    log.debug("keyToTable", "GOt value for %s\\%s = %s", key, name, table.toCsv(temp[name]["value"]) or "nil")
                else
				    log.debug("keyToTable", "GOt value for %s\\%s = %s", key, name, tostring(temp[name]["value"]) or "nil")
                end
				return { [key .. name] = temp[name]["value"] }
			else
				return nil, msg
			end
		else
			log.error("keyToTable", "Error enumerating values of %s\\%s\\.  Error: %s",
				tree, key, luacom.config.last_error or tostring(result))
			return nil, luacom.config.last_error or tostring(result)
		end
	else
		log.error("keyToTable", "Error enumerating values of %s\\%s\\.  Error: %s",
			tree, key, luacom.config.last_error or "unknown")
		return nil, luacom.config.last_error or "unknown"
	end
	-- Now recursively get all subkeys
	luacom.config.last_error = nil
	local KeyNames
	result, _, _, KeyNames = svc:EnumKey(trees[tree], key, KeyNames)
	if result and  result == 0 then
		if type(KeyNames) == "table" then
			for i = 1, #KeyNames do
				keytable[KeyNames[i]] = keyToTable(svc, tree, key .. "\\" .. KeyNames[i]) or
					{ ["(Default)"] = { vtype = "REG_SZ", value = "" }}
			end
		end
	else
		log.error("keyToTable", "Error enumerating key %s\\%s\\.  Error: %s",
			tree, key, luacom.config.last_error or tostring(result))
		return nil, luacom.config.last_error or tostring(result)
	end
	return keytable
end
------------------------------------------------------------------
------------------------------------------------------------------
-- writeValue
-- Writes a new value to to a key value.  Returns the new key
-- value pair.
-- Parameters:
--    svc - the LuaCom object to use
--    tree - the registry key (hive) to check
--    key - the key to operate on
--    value - the value to write
--    vtype - the type of the value.  If not specifed, the current
--      type will be used.  If not specified and this is a new
--      keyvalue the default of REG_SZ will be used.
-- Returns: the resulting table or nil plus error message if error
------------------------------------------------------------------
function writeValue(svc, tree, key, value, vtype)
	if not key or not svc or not tree or not value then
		log.error("keyExists", "nil request received")
		return nil
	end
	local result
	-- See if this is a value, key or doesn't exist
	local keytype = keyExists(svc, tree, key)
	-- Prep for later by splitting the value name from the kwy
	local _, _, name = string.find(key, "([^\\]*)$")
	key = string.gsub(key, "[^\\]*$", "")
	if not keytype then
		-- Key/value doesn't exist.  Does the parent key?
		keytype = keyExists(svc, tree, key)
		if keytype then
			if keytype == "value" then
				-- This should never happen
				return string.format("Illegal key.  Parent key %s is already a value", key)
			end
			log.trace("writeValue", "Key %s\\%s\\ already exists", tree, key)
		else
			-- Create the key
			log.trace("writeValue", "Creating key %s\\%s\\", tree, key)
			local pcall_result, temp = pcall(svc.CreateKey, svc, trees[tree], key)
			if not temp or temp ~= 0 then
				log.error("writeValue", "Error creating key %s\\%s\\", tree, key)
				return string.format("Error creating key %s\\%s\\", tree, key)
			end
		end
	else
		if keytype ~= "value" then return "Can't change the value of a key" end
	end
	-- If we are here the parent key exists and we have the value name split out
	-- let's try to find out what type the value is (if it exists)
	local temp = keyToTable(svc, tree, key)
	local cur_type = "REG_SZ"
	if type(temp) == "table" then
		if temp[name] then cur_type = temp[name].vtype end
	else
		log.error("writeValue", "getting type of %s\\%s\\%s", tree, key, name)
		return "Error getting current type"
	end
	-- Allow for overwriting the type
	cur_type = vtype or cur_type
	-- Now we have the type.  We can actually write the value.
	local pcall_result, write_res
	if cur_type == "REG_SZ" then
		pcall_result, write_res = pcall(svc.SetStringValue, svc, trees[tree], key, name, tostring(value))
	elseif cur_type == "REG_EXPAND_SZ" then
		pcall_result, write_res = pcall(svc.SetExpandedStringValue, svc, trees[tree], key, name, tostring(value))
	elseif cur_type == "REG_BINARY" then
	--[[ THIS IS THE RIGHT WAY TO DO IT BUT LUACOM HAS A BUG THAT INSERTS A 00 AT THE BEGINNING OF THE ARRAY
		-- Need to convert the value from a space separated set of numbers to an array
		temp = {}
		for w in string.gmatch(value, "%S+") do
			table.insert(temp,tonumber(w, 16))
		end
		-- A little weirdness with LuaCOM - we need to pad this array with and extra value
		table.insert(temp, 0)
		pcall_result, write_res = pcall(svc.SetBinaryValue, svc, trees[tree], key, name, temp)
	--]]
		-- A hack to write the binary value using a cscript
		local iValues = "Array("
		for w in string.gmatch(value, "%S+") do
			iValues = iValues .. tonumber(w, 16) .. ","
		end
		iValues = string.gsub(iValues,",$", "%)")
		local cmd1 = 'Set objReg = GetObject("winmgmts:{impersonationLevel=impersonate}!//./root/default:StdRegProv" )'
		local cmd2 = 'objReg.SetBinaryValue(' .. trees[tree] .. ',"' .. key .. '","' .. name .. '",' .. iValues .. ')'
		local wsh =  luacom.CreateObject('ScriptControl')
		if wsh then
			wsh.Language = "VBScript"
			pcall_result = pcall(wsh.ExecuteStatement, wsh, cmd1)
			pcall_result, write_res = pcall(wsh.Eval, wsh, cmd2)
		else
			log.error("writeValue", "Error creating wscript shell object")
			return string.format("Error creating wscript shell object")
		end
	elseif cur_type == "REG_DWORD" then
		pcall_result, write_res = pcall(svc.SetDWORDValue, svc, trees[tree], key, name, tonumber(value))
	else
		log.error("writeValue", "Unsupported type: %s", cur_type)
		return string.format("Unsupported type: %s", cur_type)
	end
	if not write_res or write_res ~= 0 then
		log.error("writeValue", "Error setting %s\\%s%s to %s.  Error: %s", tree, key, name, value, write_res or "unknown")
		return string.format("Error setting %s\\%s%s to %s.  Error: %s", tree, key, name, value, write_res or "unknown")
	end
	pcall_result, write_res = pcall(keyToTable, svc, tree, key .. "\\" .. name)
	return { [tree .. "\\" .. key .. name] = write_res }
end
------------------------------------------------------------------
------------------------------------------------------------------
-- keyExists
-- Returns "key", "value" if input is an existing key or value
--          return nil is the key or value does not exist
-- Parameters:
--    svc - the LuaCom object to use
--    tree - the registry key (hive) to check
--    key - the key to operate on
-- Returns: the resulting table or nil plus error message if error
------------------------------------------------------------------
function keyExists(svc, tree, key)
	if not key or not svc or not tree then
		log.error("keyExists", "nil request received")
		return nil
	end
	local result
	-- Get any values that may exist
	local ValueNames, ValueTypes
	log.trace("keyExists", "Reading key %s\\%s\\", tree, key)
	luacom.config.last_error = nil
	result, _, _, ValueNames, ValueTypes = svc:EnumValues(trees[tree], key, ValueNames, ValueTypes)
	if result then
		if result == 0  then
			-- This is a key since its values were successfully enumerated
			return "key"
		elseif result == 2 then
			-- Check to see if this is a value.  Check it by
			-- chopping off the last part of the key variable to get all the
			-- key values and then select the one we are looking for
			local _, _, name = string.find(key, "([^\\]*)$")
			key = string.gsub(key, "[^\\]*$", "")
			log.debug("keyExists", "Trying key %s with value name = %s", key, name)
			result, _, _, ValueNames, ValueTypes = svc:EnumValues(trees[tree], key, ValueNames, ValueTypes)
			if result and result == 0 and type(ValueNames) == "table" then
				for i = 0, #ValueNames do
					-- check to see if it is a value
					if ValueNames[i] == name then return "value" end
				end
			end
		end
	end
	-- Not a key or value (or there was some other error)
	return nil, luacom.config.last_error or tostring(result)
end
------------------------------------------------------------------
