--
-- modbus_handler.lua
-- Provides RESTful interface to Modbus script
--
--
-- Developed by ThingWorx
-- (c)2012 ThingWorx
--
--  Configuration Items
--
--   host or serial_port = <hostname or ip address> or <comX or ttySX>
--		One of either an TCP/IP hostname/address or a serial port
--		identifier.  Automatically switches to using Modbus TCP or Modbus
--		RTU depending on the entry.
--
--  Optional Configuration Items
-- 		port - if a host name is specified the TCP port to use.  Default = 502
-- 		baud_rate - if a serial port is selected, the baud rate to use.  Defauult=9600
--		data_bits - if a serial port is selected, the number of data bits to use.  Defauult=8
--		parity - if a serial port is selected, the parity to use.  Defauult=None
--		stop_bits - if a serial port is selected, the number of stop bits to use. Defauult=None
--		flow_control - if a serial port is selected, the flow control use.  Defauult=None
--		read_timeout - in milliseconds.  Defaults to 50 for ModbusRTU and 1 for Modbus TCP
--		response_timeout - in milliseconds.  Defaults to 5000.
--
--
-- Usage
--		GET /coil/<slave_id>/<coil_number>
--			Retrieves the status of a single coil.  Returns JSON VTQ
--			{ "value":<coil_status>, "time":<time_stamp>, "quality":"GOOD"}
--
--		GET /discrete_input/<slave_id>/<input_number>
--			Retrieves the status of a single discrete input.  Returns JSON VTQ
--			{ "value":<input_status>, "time":<time_stamp>, "quality":"GOOD"}
--
--		GET /holding_register/<slave_id>/<register>[?format=<format>&offset=<offset>&scale=<scale>&length=<length>]
--			Retrieves the status of a single holding register value.  Returns JSON VTQ
--			{ "value":<converted_value>, "time":<time_stamp>, "quality":"GOOD"}
--			Note that this may actually read more than one register if the format
--			specifier requires multiple registers (e.g. Int32).  Valid formats are
--			Int16, Uint16, Int32, Uint32, Float, DOuble, MaskedBitmap16, MaskedBitmap32, Timestamp, String.
--			Default format is Int16. Numeric values may be scaled using the "scale" and "offset" params.
--			Bitmasks return a Uin16 or Uint32 value.  Mask and length are set using the optional
--			"offset" and "length" paramters which default to 0 and 1 respectively.
--			Strings require a length parameter to specifiy the maximum length of the string. Default is 10.
--
--		GET /input_register/<slave_id>/<register>[?format=<format>&offset=<offset>&scale=<scale>&length=<length>]
--			Retrieves the status of a single input register value.  Returns JSON VTQ
--			{ "value":<converted_value>, "time":<time_stamp>, "quality":"GOOD"}
--			Note that this may actually read more than one register if the format
--			specifier requires multiple registers (e.g. Int32).  Valid formats are
--			Int16, Uint16, Int32, Uint32, Float, DOuble, MaskedBitmap16, MaskedBitmap32, Timestamp, String.
--			Default format is Int16. Numeric values may be scaled using the "scale" and "offset" params.
--			Bitmasks return a Uin16 or Uint32 value.  Mask and length are set using the optional
--			"offset" and "length" paramters which default to 0 and 1 respectively.
--			Strings require a length parameter to specifiy the maximum length of the string. Default is 10.
--
--		PUT /coil/<slave_id>/<coil_number>
--          Sets a coil to the value specfied in the post_data.
--			Post data must be JSON in the form {"value":<value>}
--
--		PUT /holding_register/<slave_id>/<register>
--          Sets a single holding register to the value specfied in the post_data.
--			Post data must be JSON in the form {"value":<value>}.  Data must be
--			a raw 16 bit value.  It is NOT formatted in any way.
--
--		POST /read_multiple_discretes
--			Retrieves the status of multiple contiguous coils or discrete inputs.
--			Returns JSON array of boolean vtq
--			Post data input to the service includes {"slave":<slave_id>,"address": <base address>,
--			"count":<number of contiguous discretes to read>}
--
--		POST /read_multiple_registers
--			Retrieves a vtq array of multiple contiguous holding or input registers.
--			Returns JSON array of formatted data
--			Post data input to the service includes {"slave":<slave_id>,"address": <starting register>,
--			"count":<number of regsiters to read>, "formats":{ JSON Arrary of format specifiers}}
--			Because of differening sizes, the number of formats in the format specifier array may not
--			be the same as the number of registers to read.  Valid format values are:
--			Int16, Uint16, Int32, Uint32, Float, Double, MaskedBitmap16, MaskedBitmap32, Timestamp, String
--			Default format is Int16. Numeric values may be scaled using the "scale" and "offset" params.
--			Bitmasks return a Uin16 or Uint32 value.  Mask and length are set using the optional
--			"offset" and "length" paramters which default to 0 and 1 respectively.
--			Example: {{"Int16", 2, 1}, "Uint16", "Int32", "Uint32", {"String",6}, {"MaskedBitmap32", 23, 2}}
--
--		POST /write_multiple_registers
--			Writes multiple contiguous holding registers.
--			Post data input to the service includes
--			{"slave":<slave_id>,"address": <starting register>, "data":[<array of unsigned 16 bit numbers]}
--
------------------------------------------------------------------
------------------------------------------------------------------
-- A function to load the tw_* libraries.  THIS IS TO BE USED WHEN
-- RUNNING IN A STANDALONE DEBUGGER ONLY.  DO NOT CALL THIS FUNCTION
-- WHEN RUNNING IN THE LUA SCRIPT RESOURCE.
-- Make sure to set you path to the proper value
function init_tw_library()
-- function to load a DLL
local path = os.getenv("USERPROFILE") .. '\\My Documents\\_ThingWorx\\Platform\\EdgeConnectivity\\extensions\\luaScripts\\debug_tw_library\\tw_library.dll'
print (path)
local f = assert(package.loadlib(path, "luaopen_tw_library"))
f()
package.path = os.getenv("USERPROFILE") .. '\\My Documents\\_ThingWorx\\Platform\\EdgeConnectivity\\extensions\\luaScripts\\lua_libs\\?.lua;' .. package.path
package.cpath = os.getenv("USERPROFILE") .. '\\My Documents\\_ThingWorx\\Platform\\EdgeConnectivity\\extensions\\luaScripts\\lua_clibs\\?.dll;' .. package.cpath
tw_logger.force("lua", "loaded tw_library")
tw_debugging = true
end
--------------------------------------------------------------------

mbFormatLength = {
	-- This is REGISTER count, not byte count
	Int16 = 1,
	Uint16 = 1,
	Int32 = 2,
	Uint32 = 2,
	Float = 2,
	Double = 4,
	String = 10,
	MaskedBitmap16 = 1,
	MaskedBitmap32 = 2,
	Timestamp = 5,
}

------------------------------------------------------------------
-- validate_req
-- Validates generic information for a read
-- Parameters:
--    required_method - a string containing the required method
--    method - a string containing the method requested (GET, POST, DELETE)
--    path - an array of strings containing the HTTP path specified in the request "/my/path/to/data"
--    headers - a table of HTTP header attribute/value pairs in the request, can be nil
-- Returns: ResponseCode, headers, ResultString, SlaveId, Register
------------------------------------------------------------------
function validate_req(required_method, method, path, headers)
    -- Make sure we have a modbus driver
	if not mb then
	    -- Try to create our modbus driver (again)
		mb = require("modbus").init(config)
		if not mb then return 500, headers, "No Modbus driver detected" end
	end
	-- Set our content type, default to json
	local contentType = "application/json"
	if headers and headers["Accept"] then
		if string.find(headers["Accept"],"application/json") then
			-- do nothing
		elseif string.find(headers["Accept"],"text/xml") then
			contentType = "text/xml"
		elseif string.find(headers["Accept"],"text/plain") then
			contentType = "text/plain"
		end
	end
	headers["content-type"] = contentType
	-- Now process the request
	local err
	local result = {}
	if method == required_method then
		if required_method == "POST" then
			return 200, headers, "OK"
		-- Check for a valid slave and register
		elseif (path[2] and path[3]) then
		    local slave = tonumber(path[2])
			local reg = tonumber(path[3])
			if slave and slave < 247 and reg and reg < 65535 then
				-- This is a validated request.  Calling function will do the actual read
					return 200, headers, "OK", math.floor(slave), math.floor(reg)
			else
			    headers["content-type"] = "text/plain"
				return 400, headers, "Invalid slave: " .. (path[2] or "nil") .. " or register: " .. (path[3] or "nil")
			end
		else
			headers["content-type"] = "text/plain"
			return 400, headers, "Must specify a slave id and register"
		end
	else
		code = 405
		headers["content-type"] = "text/plain"
		return 405, headers, method .. " not supported for this URI"
	end
end
------------------------------------------------------------------
------------------------------------------------------------------
-- read_coil_status
-- Reads the status of a single coil
-- Parameters:
--    method - a string containing the method requested (GET, POST, DELETE)
--    path - an array of strings containing the HTTP path specified in the request "/my/path/to/data"
--    headers - a table of HTTP header attribute/value pairs in the request, can be nil
--    query - a table of attribute/value pairs in any query string ("?myattribute=myvalue"), can be nil
--    postdata - a table of attribute/value pairs in post data ("myattribute=myvalue"), can be nil
-- Returns: ResponseCode, Headers, Value
------------------------------------------------------------------
function read_coil_status(method, path, headers, query, postdata)
    --
	local code, result, slaveid, register
	code, headers, result, slaveid, register = validate_req("GET", method, path, headers)
	if code ~= 200 then
		return code, headers, result
	end
	-- Now execute the request
	tw_mutex.lock()
	local res, err = mb:executeFunction(slaveid, "READ_COIL_STATUS", register, 1)
	tw_mutex.unlock()
	if type(res) == "table" and type(res[1]) == "string" then
		local vtq = {}
		vtq.quality = "GOOD"
		vtq.value = res[1]:byte(1)
		vtq.time = os.time() * 1000
		return 200, headers, table.toContentTypeString(vtq, headers["content-type"], query)
	else return 500, headers, err or "unknown"
	end
end
------------------------------------------------------------------
------------------------------------------------------------------
-- read_discrete_status
-- Reads the status of a single discrete input
-- Parameters:
--    method - a string containing the method requested (GET, POST, DELETE)
--    path - an array of strings containing the HTTP path specified in the request "/my/path/to/data"
--    headers - a table of HTTP header attribute/value pairs in the request, can be nil
--    query - a table of attribute/value pairs in any query string ("?myattribute=myvalue"), can be nil
--    postdata - a table of attribute/value pairs in post data ("myattribute=myvalue"), can be nil
-- Returns: ResponseCode, Headers, Value
------------------------------------------------------------------
function read_discrete_status(method, path, headers, query, postdata)
    --
	local code, result, slaveid, register
	code, headers, result, slaveid, register = validate_req("GET", method, path, headers)
	if code ~= 200 then
		return code, headers, result
	end
	-- Now execute the request
	tw_mutex.lock()
	local res, err = mb:executeFunction(slaveid, "READ_DISCRETE_STATUS", register, 1)
	tw_mutex.unlock()
	if type(res) == "table" and type(res[1]) == "string" then
		local vtq = {}
		vtq.quality = "GOOD"
		vtq.value = res[1]:byte(1)
		vtq.time = os.time() * 1000
		return 200, headers, table.toContentTypeString(vtq, headers["content-type"], query)
	else return 500, headers, err or "unknown"
	end
end
------------------------------------------------------------------
------------------------------------------------------------------
-- read_multiple_discretes
-- Reads the status of multiple contiguous discrete inputs
-- Parameters:
--    method - a string containing the method requested (GET, POST, DELETE)
--    path - an array of strings containing the HTTP path specified in the request "/my/path/to/data"
--    headers - a table of HTTP header attribute/value pairs in the request, can be nil
--    query - a table of attribute/value pairs in any query string ("?myattribute=myvalue"), can be nil
--    postdata - a table of attribute/value pairs in post data ("myattribute=myvalue"), can be nil
-- Returns: ResponseCode, Headers, Value
------------------------------------------------------------------
function read_multiple_discretes(method, path, headers, query, postdata)
	local code, result
	code, headers, result = validate_req("POST", method, path, headers)
	if code ~= 200 then
		return code, headers, result
	end
	local input = json.decode(postdata.content or {})
	if input and tonumber(input.slave) and
		tonumber(input.address) and tonumber(input.count) then
		if tonumber(input.slave) < 247 and tonumber(input.address) < 65535 then
			local slave, addr, count = tonumber(input.slave), tonumber(input.address), tonumber(input.count)
			-- Now execute the request
			local func = "READ_DISCRETE_STATUS"
			if tonumber(input.address) < 10000 then func = "READ_COIL_STATUS" end
			tw_mutex.lock()
			local res, err = mb:executeFunction(slave, func, addr, count)
			tw_mutex.unlock()
			if type(res) == "table" then
				-- COnvert to individual bits
				local time = os.time() * 1000
				local vtq = {}
				for x = 1, math.ceil(count/8) do
					if x > string.len(res[1]) then break end
					local byte = res[1]:byte(x)
					for y = 0, 7, 1 do
						local tmp = {}
						tmp.time = time
						tmp.value = math.floor(byte%2)
						tmp.quality = "GOOD"
						table.insert(vtq, tmp)
						byte = math.floor(byte/2)
						if #vtq == count then break end
					end
				end
				return 200, headers, table.toContentTypeString(vtq, headers["content-type"], query)
			else return 500, headers, err or "unknown"
			end
		else
			headers["content-type"] = "text/plain"
			return 400, headers, "Invalid slave or address"
		end
	else
		headers["content-type"] = "text/plain"
		return 400, headers, "Must specify a slave id, address, and count"
		end
end
------------------------------------------------------------------
------------------------------------------------------------------
-- read_holding_register
-- Reads and formats a single holding register
-- Parameters:
--    method - a string containing the method requested (GET, POST, DELETE)
--    path - an array of strings containing the HTTP path specified in the request "/my/path/to/data"
--    headers - a table of HTTP header attribute/value pairs in the request, can be nil
--    query - a table of attribute/value pairs in any query string ("?myattribute=myvalue"), can be nil
--    postdata - a table of attribute/value pairs in post data ("myattribute=myvalue"), can be nil
-- Returns: ResponseCode, Headers, Value
------------------------------------------------------------------
function read_holding_register(method, path, headers, query, postdata)
    --
	local code, result, slaveid, register
	code, headers, result, slaveid, register = validate_req("GET", method, path, headers)
	if code ~= 200 then
		return code, headers, result
	end
	-- check our query string for a format
	local fmt = { "Int16" }
	local count = 1
	if query then
		if query.format and mbFormatLength[query.format] then fmt = {query.format} end
		count = mbFormatLength[query.format or "Int16"]
		-- Check to see if there are any additoinal format specifiers
		if fmt[1] == "Int16" or fmt[1] == "Int32" or fmt[1] == "Uint16" or fmt[1] == "Uint32"
		or fmt[1] == "Float"  or fmt[1] == "Double" then
			if query.scale then fmt[2] = query.scale end
			if query.offset then fmt[3] = query.offset end
		end
		if fmt[1] == "MaskedBitmap16" or fmt[1] == "MaskedBitmap32" then
			if query.offset then fmt[2] = query.offset end
			if query.length then fmt[3] = query.length end
		end
		if fmt[1] == "String" then
			fmt[2] = 10
			count = 5
			if tonumber(query.length) then
				fmt[2] = query.length
				count = math.ceil(tonumber(query.length)/2)
			end

		end
	end
	-- Now execute the request
	tw_mutex.lock()
	local res, err = mb:executeFunction(slaveid, "READ_HOLDING_REGISTERS", register, count, nil, {fmt})
	tw_mutex.unlock()
	if type(res) == "table" and (type(res[1]) == "string" or type(res[1]) == "number") then
		local vtq = {}
		vtq.quality = "GOOD"
		vtq.value = res[1]
		vtq.time = os.time() * 1000
		return 200, headers, table.toContentTypeString(vtq, headers["content-type"], query)
	else return 500, headers, err or "unknown"
	end
end
------------------------------------------------------------------
------------------------------------------------------------------
-- read_input_register
-- Reads and formats a single input register
-- Parameters:
--    method - a string containing the method requested (GET, POST, DELETE)
--    path - an array of strings containing the HTTP path specified in the request "/my/path/to/data"
--    headers - a table of HTTP header attribute/value pairs in the request, can be nil
--    query - a table of attribute/value pairs in any query string ("?myattribute=myvalue"), can be nil
--    postdata - a table of attribute/value pairs in post data ("myattribute=myvalue"), can be nil
-- Returns: ResponseCode, Headers, Value
------------------------------------------------------------------
function read_input_register(method, path, headers, query, postdata)
    --
	local code, result, slaveid, register
	code, headers, result, slaveid, register = validate_req("GET", method, path, headers)
	if code ~= 200 then
		return code, headers, result
	end
	-- check our query string for a format
	local fmt = { "Int16" }
	local count = 1
	if query then
		if query.format and mbFormatLength[query.format] then fmt = {query.format} end
		count = mbFormatLength[query.format or "Int16"]
		-- Check to see if there are any additoinal format specifiers
		if fmt[1] == "Int16" or fmt[1] == "Int32" or fmt[1] == "Uint16" or fmt[1] == "Uint32"
		or fmt[1] == "Float"  or fmt[1] == "Double" then
			if query.scale then fmt[2] = query.scale end
			if query.offset then fmt[3] = query.offset end
		end
		if fmt[1] == "MaskedBitmap16" or fmt[1] == "MaskedBitmap32" then
			if query.offset then fmt[2] = query.offset end
			if query.length then fmt[3] = query.length end
		end
		if fmt[1] == "String" then
			fmt[2] = 10
			count = 5
			if tonumber(query.length) then
				fmt[2] = query.length
				count = math.ceil(tonumber(query.length)/2)
			end

		end
	end
	-- Now execute the request
	tw_mutex.lock()
	local res, err = mb:executeFunction(slaveid, "READ_INPUT_REGISTERS", register, count, nil, {fmt})
	tw_mutex.unlock()
	if type(res) == "table" and (type(res[1]) == "string" or type(res[1]) == "number") then
		local vtq = {}
		vtq.quality = "GOOD"
		vtq.value = res[1]
		vtq.time = os.time() * 1000
		return 200, headers, table.toContentTypeString(vtq, headers["content-type"], query)
	else return 500, headers, err or "unknown"
	end
end
------------------------------------------------------------------
------------------------------------------------------------------
-- read_multiple_registers
-- Reads the status of multiple contiguous registers
-- Parameters:
--    method - a string containing the method requested (GET, POST, DELETE)
--    path - an array of strings containing the HTTP path specified in the request "/my/path/to/data"
--    headers - a table of HTTP header attribute/value pairs in the request, can be nil
--    query - a table of attribute/value pairs in any query string ("?myattribute=myvalue"), can be nil
--    postdata - a table of attribute/value pairs in post data ("myattribute=myvalue"), can be nil
-- Returns: ResponseCode, Headers, Value
------------------------------------------------------------------
function read_multiple_registers(method, path, headers, query, postdata)
	local code, result
	code, headers, result = validate_req("POST", method, path, headers)
	if code ~= 200 then
		return code, headers, result
	end
	local input = json.decode(postdata.content or {})
	if input and tonumber(input.slave) and
		tonumber(input.address) and tonumber(input.count) then
		if tonumber(input.slave) < 247 and tonumber(input.address) < 65535 then
			local slave, addr, count = tonumber(input.slave), tonumber(input.address), tonumber(input.count)
			-- Default to Int16 formats
			local reg_array = {}			
			if not input.formats then
				input.formats = {}
				for i = 1, count do
					table.insert(reg_array, addr + (i - 1))
					table.insert(input.formats,"Int16")
				end
			end
			-- Now check for other formats.  If they exist we might
			-- need to adjust our count
			if type(input.formats) == "table" then
				count = 0
				local tmp
				for i=1, #input.formats do
					if type(input.formats[i]) == "string" then tmp = input.formats[i]
					else tmp = input.formats[i][1] end
					if mbFormatLength[tmp] then
						table.insert(reg_array, addr + count)
						if tmp == "String" then count = count + (tonumber(input.formats[i][2]) or mbFormatLength[tmp])
						else count = count + mbFormatLength[tmp] end
					else return 400, headers, "Invalid format specifier: " .. tmp or "nil" end
				end
			end
			-- Execute the request
			local func = "READ_HOLDING_REGISTERS"
			if addr < 40000 then func = "READ_INPUT_REGISTERS" end
			tw_mutex.lock()
			local res, err = mb:executeFunction(slave, func, addr, count, nil, input.formats)
			tw_mutex.unlock()
			if type(res) == "table" then
				local vtq = {}
				local time = os.time() * 1000
				for k,v in ipairs(res) do
					local tmp = {}
					tmp.value = v
					tmp.quality = "GOOD"
					tmp.time = time
					if k <= #reg_array then tmp.register = reg_array[k] end
					table.insert(vtq,tmp)
				end
				return 200, headers, table.toContentTypeString(vtq, headers["content-type"], query)
			else 
				log.error("read_multiple_registers","Invalid return type")
				return 500, headers, err or "unknown"
			end
		else
			headers["content-type"] = "text/plain"
			return 400, headers, "Invalid slave or address"
		end
	else
		headers["content-type"] = "text/plain"
		return 400, headers, "Must specify a slave id, address, and count"
	end
end
------------------------------------------------------------------
------------------------------------------------------------------
-- write_multiple_registers
-- Writes multiple holding registers.  Input is JSON array of 16 bit values.
-- Parameters:
--    method - a string containing the method requested (GET, POST, DELETE)
--    path - an array of strings containing the HTTP path specified in the request "/my/path/to/data"
--    headers - a table of HTTP header attribute/value pairs in the request, can be nil
--    query - a table of attribute/value pairs in any query string ("?myattribute=myvalue"), can be nil
--    postdata - a table of attribute/value pairs in post data ("myattribute=myvalue"), can be nil
-- Returns: ResponseCode, Headers, Value
------------------------------------------------------------------
function write_multiple_registers(method, path, headers, query, postdata)
	--
	local code, result, slaveid, register
	code, headers, result, slaveid, register = validate_req("POST", method, path, headers)
	if code ~= 200 then
		return code, headers, result
	end
	local input = json.decode(postdata.content or {})
	if input and tonumber(input.slave) and tonumber(input.address) then
		if tonumber(input.slave) < 247 and tonumber(input.address) < 65535 then
			local slave, addr = tonumber(input.slave), tonumber(input.address)
			local tmp = input.data
			if type(tmp) ~= "table" then
				return 400, headers, 'POST must have valid JSON array of unsigned 16 bit numbers'
			end
			local val = {}
			for i = 1, #tmp do
				if not tonumber(tmp[i]) or tonumber(tmp[i]) < 0 or tonumber(tmp[i]) > 65535 then
					return 400, headers, 'POST must have valid JSON array of unsigned 16 bit numbers'
				end
				table.insert(val,math.floor(tmp[i]/256))
				table.insert(val,tmp[i]%256)
			end
			val = string.char(unpack(val))
			tw_mutex.lock()
			local res, err = mb:executeFunction(slave, "WRITE_MULTIPLE_REGISTERS", addr, #tmp, val)
			tw_mutex.unlock()
			if type(res) == "table" and res[1] == 0 then return 200, headers, "OK"
			else return 500, headers, err or "unknown" end
		else
			headers["content-type"] = "text/plain"
			return 400, headers, "Invalid slave or address"
		end
	else
		headers["content-type"] = "text/plain"
		return 400, headers, "Must specify a slave id, address, and count"
	end
end
------------------------------------------------------------------
------------------------------------------------------------------
-- coil
-- Reads or writes a single coil
-- Parameters:
--    method - a string containing the method requested (GET, POST, DELETE)
--    path - an array of strings containing the HTTP path specified in the request "/my/path/to/data"
--    headers - a table of HTTP header attribute/value pairs in the request, can be nil
--    query - a table of attribute/value pairs in any query string ("?myattribute=myvalue"), can be nil
--    postdata - a table of attribute/value pairs in post data ("myattribute=myvalue"), can be nil
-- Returns: ResponseCode, Headers, Value
------------------------------------------------------------------
function coil(method, path, headers, query, postdata)
    -- If it is a PUT handle it, otherwise pass it off
	if method ~= "PUT" then return read_coil_status(method, path, headers, query, postdata) end
	--
	local code, result, slaveid, register
	code, headers, result, slaveid, register = validate_req("PUT", method, path, headers)
	if code ~= 200 then
		return code, headers, result
	end
	-- Now execute the PUT request
	if not postdata or type(postdata.content) ~= "string" then
		return 400, headers, 'PUT must have valid JSON. {"value":<1|0>}'
	end
	local tmp = json.decode(postdata.content)
	if type(tmp) ~= "table" or not tmp.value then
		return 400, headers, 'PUT must have valid JSON. {"value":<1|0>}'
	end
	local val = tonumber(tmp.value)
	if not val or (val ~= 1 and val ~= 0) then
		return 400, headers, 'PUT must have valid JSON. {"value":<1|0>}'
	end
	if val == 1 then val = string.char(0xff, 0x00)
	else val = string.char(0x00, 0x00) end
	tw_mutex.lock()
	local res, err = mb:executeFunction(slaveid, "WRITE_SINGLE_COIL", register, 1, val)
	tw_mutex.unlock()
	if type(res) == "table" and res[1] == 0 then return 200, headers, "OK"
	else return 500, headers, err or "unknown"
	end
end
------------------------------------------------------------------
------------------------------------------------------------------
-- holding_register
-- Reads or writes a single holding register
-- Parameters:
--    method - a string containing the method requested (GET, POST, DELETE)
--    path - an array of strings containing the HTTP path specified in the request "/my/path/to/data"
--    headers - a table of HTTP header attribute/value pairs in the request, can be nil
--    query - a table of attribute/value pairs in any query string ("?myattribute=myvalue"), can be nil
--    postdata - a table of attribute/value pairs in post data ("myattribute=myvalue"), can be nil
-- Returns: ResponseCode, Headers, Value
------------------------------------------------------------------
function holding_register(method, path, headers, query, postdata)
    -- If it is a PUT handle it, otherwise pass it off
	if method ~= "PUT" then return read_holding_register(method, path, headers, query, postdata) end
	--
	local code, result, slaveid, register
	code, headers, result, slaveid, register = validate_req("PUT", method, path, headers)
	if code ~= 200 then
		return code, headers, result
	end
	-- Now execute the PUT request
	if not postdata or type(postdata.content) ~= "string" then
		return 400, headers, 'PUT must have valid JSON. {"value":<number>}'
	end
	local tmp = json.decode(postdata.content)
	if type(tmp) ~= "table" or not tmp.value then
		return 400, headers, 'PUT must have valid JSON. {"value":<number>}'
	end
	local val = tonumber(tmp.value)
	if not val or val < 0 or val > 65535 then
		return 400, headers, 'PUT must have valid JSON. {"value":<number>}'
	end
	val = string.char(math.floor(val/256), val%256)
	tw_mutex.lock()
	local res, err = mb:executeFunction(slaveid, "WRITE_SINGLE_REGISTER", register, 1, val)
	tw_mutex.unlock()
	if type(res) == "table" and res[1] == 0  then return 200, headers, "OK"
	else return 500, headers, err or "unknown" end
end
------------------------------------------------------------------
------------------------------------------------------------------
------------------------------------------------------------------
-- Main Loop
------------------------------------------------------------------
-- Initialiize the tw_library if we are running outside
-- of the luaScriptResource for testing putposes
--------------------------------
---- For debugging purposes ----
--------------------------------
if (p_data == nil) then
	print("Starting debug DLL")
	-- Set up our EMS host/port
	p_data = {}
	p_data.rap_host = "localhost"
	p_data.rap_port = "8000"
	-- Load the debug library
	init_tw_library()
	tw_debugging = true
	-- Properties set by luaScriptResource
	p_data.name = "modbus_handler"
	p_data.run = true
	-- Script specific properties
	p_data.host = "localhost"
	--p_data.serial_port = "COM15"
end
--------------------------------
-- Load our utility libraries
require "log"
require "stringx"
require "tablex"
json = require "json"
-- Load our modbus library
local config = {}
config.host = p_data.host
config.port = p_data.port
config.serial_port = p_data.serial_port
config.baud_rate = p_data.baud_rate
config.data_bits = p_data.data_bits
config.parity = p_data.parity
config.stop_bits = p_data.stop_bits
config.flow_control = p_data.flow_control
config.read_timeout = p_data.read_timeout or 50
config.reg_adjust = p_data.reg_adjust or 1
mb = require("modbus").init(config)


p_data.version = "1.0.0"
tw_logger.force("lua", "starting " .. p_data.name .. " script.  Version " .. p_data.version);

-- Set up some defaults
local headers = {}
headers["content-type"] = "application/json"

-- Set up our callback(s)
tw_http.registerCallback("/coil", "coil")
tw_http.registerCallback("/discrete_input", "read_discrete_status")
tw_http.registerCallback("/read_multiple_discretes", "read_multiple_discretes")
tw_http.registerCallback("/holding_register", "holding_register")
tw_http.registerCallback("/input_register", "read_input_register")
tw_http.registerCallback("/read_multiple_registers", "read_multiple_registers")
tw_http.registerCallback("/write_multiple_registers", "write_multiple_registers")

log.debug("mainloop","Starting Modbus Handler ...")
----------------------------------------------------------------
----------------------------------------------------------------
-- Loop until we are shut down
while (p_data.run == true) do
	tw_utils.psleep(500)
end
p_data.state = "Shutting down"
return ("Warning - " .. p_data.name .. " script has terminated")


