--------------------------------------------------------------------------------
-- A Shape for local rule processing
--
-- <p>
-- Configuration:
-- <ul>
--  <li><strong></strong>

--      </li>
-- <ul>
-- </p>
--------------------------------------------------------------------------------

module ("shapes.rules", thingworx.shape.extend)

p_data.rules = p_data.rules or {}
-- ------------------------------------------------------------------------------
-- Service definitions
-- ------------------------------------------------------------------------------
serviceDefinitions.UpdateRemoteRule(
  input { name="guid", baseType="STRING", description="The unique id of the rule" }, 
  input { name="rule", baseType="STRING", description="The JSON formatted rule to apply" },
  input { name="property", baseType="STRING", description="The property to apply the rule to" }, 
  output { baseType="BOOLEAN", description="Returns true if the rule was processed successfully" },
  description { "Update or adds a rule to process" }
)

serviceDefinitions.RemoveRemoteRule(
  input { name="guid", baseType="STRING", description="The unique id of the rule" }, 
  output { baseType="BOOLEAN", description="Returns true if the rule was processed successfully" },
  description { "Update or adds a rule to process" }
)

--------------------------------------------------------------------------------
-- Adds or updates a rule in the rule table
--
-- @param me A reference to the current Thing
-- @param headers The headers from the REST request invoking the service
-- @param query The query parameters from the REST request invoking the service
-- @param data Table containing service's input parameters:
--    <ul>
--      <li>data.guid The unique id of the rule</li>
--      <li>data.rule The rule to apply</li>
--      <li>data.property The property to apply the rule to</li>
--    </ul>
--
-- @return 200, true 
--
services.UpdateRemoteRule = function(me, headers, query, data)
  log.info(me.name, "Called UpdateRemoteRule")
  if data and data.rule and data.property and data.guid then
    local rule = json.decode(data.rule)
	if type(rule) == "table" then 
		tmp = {}
		tmp.property = data.property
		-- reset any triggers
		tmp.triggered = false
		-- {"filters":{"value":80.0,"fieldName":"value","type":"GT"}}
		-- "to":100.0,"from":80.0,"fieldName":"value","type":"BETWEEN"
		tmp.rule = rule.filters
		if rule.filters.type and rule.filters.type == "LIKE" or rule.filters.type == "NOTLIKE" then
		-- We want to preprocess any string match to escape any magic characters
			local match = tostring(rule.filters.value) or ""
			match = match:gsub('[%-%.%+%[%]%(%)%$%^%%%?%*]','%%%1'):gsub('%z','%%z')
			-- Now check to see if we are looking for an anchor
			rule.filters.operation = " is Exactly "
			if match:find("^%%%*") and match:find("%%%*$") then
				log.debug(me.name, "This is 'Contains' rule")
				match = match:gsub("^%%%*","", 1)
				match = match:gsub("%%%*$","")
				rule.filters.operation = " Contains "
			elseif match:find("^%%%*") then
				-- This is "Ends with"
				log.debug(me.name, "This is 'Ends with' rule")
				match = match:gsub("^%%%*","") .. "$"
				rule.filters.operation = " Ends with "
			elseif  match:find("%%%*$") then
				-- This is "Starts with"
				log.debug(me.name, "This is 'Starts with' rule")
				match = "^" .. match:gsub("%%%*$","")	
				rule.filters.operation = " Starts with "
			end	
			rule.filters.value = match
			log.debug("UpdateRemoteRule", "Updated matching string to %s", match)
		end
		tw_mutex.lock()
		p_data.rules[data.guid] = tmp
		-- Persist the rules
		tw_script.setProperties(me.name, { rules = p_data.rules } )
		tw_mutex.unlock()
		return 200, true
	else 	
		log.error(me.name,"Bad rule data: " .. data.rule)
		return 400, false
	end	
  else
	log.error(me.name, "Missing data")
    return 400, false
  end
end

--------------------------------------------------------------------------------
-- Removes a rule from the rule table
--
-- @param me A reference to the current Thing
-- @param headers The headers from the REST request invoking the service
-- @param query The query parameters from the REST request invoking the service
-- @param data Table containing service's input parameters:
--    <ul>
--      <li>data.guid The unique id of the rule</li>
--    </ul>
--
-- @return 200, true 
--
services.RemoveRemoteRule = function(me, headers, query, data)
  log.info(me.name, "Called UpdateRemoteRule")
  if data and data.guid then
    tw_mutex.lock()
	p_data.rules[data.guid] = nil
	tw_script.setProperties(me.name, { rules = p_data.rules } )
	tw_mutex.unlock()	
	return 200, true
  else
	log.error(me.name, "Missing data")
    return 400, false
  end
end

-- Evaluates a rule
--
-- @param alertId The guid of the rule
-- @param property The name of the property
-- @param rule The JSON formatted rule
-- @param value The current value of the property
-- @param isTriggered The current trigger state of the rule
--
-- @return bool True if the rule triggered, false if not
function evaluateRule(alertId, property, currentValue, rule, isTriggered)
	if not property or not currentValue or not rule or type(rule) ~= "table" then
		log.error("evaluateRule", "Missing or invalid parameter. %s, %s, %s", property, currentValue, table.tostring(rule))
		return
	end	
	local newTrigger = false
	log.debug("evaluateRule", "Checking rule %s on property", rule.type,  property)
	local cvNumber = tonumber(currentValue)
	local valNumber = tonumber(rule.value)
	local highNumber = tonumber(rule.to)
	local lowNumber = tonumber(rule.from)
	local cvString = tostring(currentValue)
	local valString = tostring(rule.value)	
	local operation = " "
	local target = valString
	local alertData = { DeviceName = p_data.name, Property = property, Timestamp = os.time() * 1000 }
	log.debug("evaluateRule", "Checking if %s is %s than %s", cvNumber,  rule.type, valNumber )
	if rule.type == "GT" then
		operation = " Greater Than "
		if cvNumber ~= nil and valNumber ~= nil then
			log.debug("evaluateRule", "Checking if %s is %s %s", cvNumber,  operation, valNumber )
			if cvNumber > valNumber then
				newTrigger = true
			end
		else
			log.error("evaluateRule", "Missing value in %s rule", rule.type)
		end
	elseif rule.type == "GTE" then
		operation = " Greater Than or Equal To "
		if cvNumber ~= nil and valNumber ~= nil then
			log.debug("evaluateRule", "Checking if %s is %s %s", cvNumber,  operation, valNumber )
			if cvNumber >= valNumber then
				newTrigger = true
			end
		else
			log.error("evaluateRule", "Missing value in %s rule", rule.type)
		end
	elseif rule.type == "LT" then
		operation = " Less Than "
		if cvNumber ~= nil and valNumber ~= nil then
			log.debug("evaluateRule", "Checking if %s is %s %s", cvNumber,  operation, valNumber )
			if cvNumber < valNumber then
				newTrigger = true
			end
		else
			log.error("evaluateRule", "Missing value in %s rule", rule.type)
		end
	elseif rule.type == "LTE" then
		operation = " Less Than or Equal To "
		if cvNumber ~= nil and valNumber ~= nil then
			log.debug("evaluateRule", "Checking if %s is %s %s", cvNumber,  operation, valNumber )
			if cvNumber <= valNumber then
				newTrigger = true
			end
		else
			log.error("evaluateRule", "Missing value in %s rule", rule.type)
		end
	elseif rule.type == "EQ" and type(rule.value) ~= "boolean" then
		operation = " Equal To "
		if cvNumber ~= nil and valNumber ~= nil then
			log.debug("evaluateRule", "Checking if %s is %s %s", cvNumber,  operation, valNumber )
			if cvNumber == valNumber then
				newTrigger = true
			end
		else
			log.error("evaluateRule", "Missing value in %s rule", rule.type)
		end
	elseif rule.type == "NE" then
		operation = " Not Equal To "
		if cvNumber ~= nil and valNumber ~= nil then
			log.debug("evaluateRule", "Checking if %s is %s %s", cvNumber,  operation, valNumber )
			if cvNumber ~= valNumber then
				newTrigger = true
			end
		else
			log.error("evaluateRule", "Missing value in %s rule", rule.type)
		end
	elseif rule.type == "BETWEEN" then
		operation = " Between "
		if cvNumber ~= nil and lowNumber ~= nil and highNumber ~= nil then
			target = lowNumber .. " and " .. highNumber
			log.debug("evaluateRule", "Checking if %s is %s %s and %s", cvNumber,  operation, lowNumber, highNumber )
			if cvNumber < highNumber and cvNumber > lowNumber then
				newTrigger = true
			end
		else
			log.error("evaluateRule", "Missing value in %s rule", rule.type)
		end
	elseif rule.type == "NOTBETWEEN" then
		operation = " Not Between "
		if cvNumber ~= nil and lowNumber ~= nil and highNumber ~= nil then
			target = lowNumber .. " and " .. highNumber
			log.debug("evaluateRule", "Checking if %s is %s %s and %s", cvNumber,  operation, lowNumber, highNumber )
			if cvNumber > highNumber or cvNumber < lowNumber then
				newTrigger = true
			end
		else
			log.error("evaluateRule", "Missing value in %s rule", rule.type)
		end
	elseif rule.type == "EQ" and type(rule.value) == "boolean" then
		operation = " is " .. rule.value .. " "
		if currentValue ~= nil then
			log.debug("evaluateRule", "Checking if %s is %s", property,  rule.value )
			if currentValue == rule.value then
				newTrigger = true
			end
		else
			log.error("evaluateRule", "Missing value in %s rule", rule.type)
		end
	elseif rule.type == "LIKE" then
		-- We have already preprocessed the matching value
		operation = rule.operation or " LIKE "
		if valString ~= nil and cvString ~= nil then
			log.debug("evaluateRule", "Checking if %s %s %s", property,  rule.operation or " LIKE ", valString )
			if cvString:find(valString) then
				newTrigger = true
			end
		else
			log.error("evaluateRule", "Missing value in %s rule", rule.type)
		end
	elseif rule.type == "NOTLIKE" then
		-- Has the string property changed
		operation = " Changed From "
		if valString ~= nil and cvString ~= nil then
			log.debug("evaluateRule", "Checking if %s is %s", property,  rule.operation or " NOTLIKE ", valString )
			if currentValue == rule.value then
				newTrigger = true
			end
		else
			log.error("evaluateRule", "Missing value in %s rule", rule.type)
		end	
	end
	if newTrigger then
		-- Only trigger if we are not already triggered
		if not isTriggered then
			alertData.AlertName = alertId
			alertData.Details = property .. " " .. cvString .. operation .. target
			log.info("evaluateRule", "Firing Event.  Details: %s", table.toString(alertData) )
			thingworx.server.fireEvent("EdgeGeneratedAlert", alertData)
		else
			log.info("evaluateRule", property .. " " .. cvString .. operation .. target .. " evalutated to true but we have already triggered" )
		end
	end		
	return newTrigger
end

ruleCache = {}

tasks.EvaluateRules = function(me, propertyName)
  -- Evaluate our rules
  log.debug("tasks.EvaluateRules", "Executing with propertyName = %s ....", propertyName or "")
  --tw_mutex.lock()
  if type(p_data.rules) == "table" then
    -- If we were passed a property in, see if we have it cached
	if propertyName and ruleCache[propertyName] then
		for k,v in pairs(ruleCache[propertyName]) do
			local r = p_data.rules[k]
			if r then
				log.debug("tasks.EvaluateRules","Found %s in ruleCache", propertyName)
				local currentValue = me.properties[r.property].value
				r.triggered = evaluateRule(k, r.property, currentValue, r.rule, r.triggered)	
				return
			else
				-- Not found. Clean up our cache
				ruleCache[propertyName][k] = nil
				log.debug("tasks.EvaluateRules","Removing rule %s for %s from ruleCache", k, propertyName)
			end
		end
	end
	for k,v in pairs(p_data.rules) do
		log.debug("tasks.EvaluateRules", "Checking rule %s ", k)
		if type(v) == "table" and v.property and v.rule then
			log.debug("tasks.EvaluateRules", "Checking property %s ", v.property)
			if me.properties[v.property] then
				if me.properties[v.property].dataChangeType == "NEVER" then
					if not propertyName then 
						me:getProperty(k)
						local currentValue = me.properties[v.property].value
						v.triggered = evaluateRule(k, v.property, currentValue, v.rule, v.triggered)
					end
				else
					if not propertyName or propertyName == v.property then
						local currentValue = me.properties[v.property].value
						v.triggered = evaluateRule(k, v.property, currentValue, v.rule, v.triggered)
						-- Update our cache
						if propertyName then 
							if not ruleCache[propertyName] then ruleCache[propertyName] = {} end
							ruleCache[propertyName][k] = 1 
						end
					end
				end
			else
				log.warn("tasks.EvaluateRules", "No handler found for %s.  Rule %s ignored.", v.property, k)
			end
		end
	end
  else
     -- No rules
	 log.debug("tasks.EvaluateRules", "No rules found")
	 return
  end
  --tw_mutex.unlock()
end


