--------------------------------------------------------------------------------
-- A Shape for running a collection of testcases.
--
-- <p>
-- Configuration:
-- <ul>
--  <li><strong>testcaseDir:</strong>
--      Path to the directory that contains the testcases to be executed.
--      </li>
--  <li><strong>testcases:</strong>
--      A table containing the names of the testcases to run. Optional. If
--      no testcases table is specified, then all testcases found in
--      testcaseDir will be executed.
--      </li>
-- </ul>
-- </p>
--------------------------------------------------------------------------------

module ("shapes.testsuite", thingworx.shape.extend)

--------------------------------------------------------------------------------
-- Default values for config parameters
-- @class table
-- @name p_data
-- @field testcaseDir Path to the directory that contains the testcases to be 
--                    executed. Defaults to 'etc/testcases'.
--

-- Add the testcaseDir to the package.path for easy loading later
if tw_utils.isWindows() then
    p_data.testcaseDir = p_data.testcaseDir or ".\\etc\\testcases"
    package.path = package.path .. ";" .. p_data.testcaseDir .. "\\?.lua"
else
    p_data.testcaseDir = p_data.testcaseDir or "etc/testcases"
    package.path = package.path .. ";" .. p_data.testcaseDir .. "/?.lua"
end

serviceDefinitions.RunTests(
  output { baseType="STRING", description="A report of the test results" },
  description { "Run all configured tests" }
)

--------------------------------------------------------------------------------
-- Runs all of the specified testcases and returns a string containing a report
-- of the test results.
--
-- @return A plain text string containing the test results
--
services.RunTests = function(me, headers, query, data)

  -- Add in all cases defined in p_data.testcaseDir if no testcases
  -- are specified in p_data.testcases.
  if not p_data.testcases then
    local files = tw_dir.listFiles(p_data.testcaseDir or ".")
    p_data.testcases = {}
    for key,val in pairs(files) do
      if not key:match("\.swp$") then -- Ignore swp files for vi person
        table.insert(p_data.testcases, key:match("(.*)\.lua"))
      end
    end
  end

  -- Execute all the specified testcases
  local result = ""

  for _,test in ipairs(p_data.testcases) do
    log.info("testsuite", "Loading testcase: %s", test)
    local case = require(test) 
    case.run(me, data)
    result = result .. "\n" .. case.report()
    case.reset()
  end

  return 200, result
end
