require "mqtt"
require "json"

module Centurysys
  class AWSIotShadowException < ::Exception
  end

  class AWSIotShadowDisconnectException < AWSIotShadowException
  end

  class AWSIotShadow
    def initialize(host: nil, port: 8883, ca_file: nil, key_file: nil,
      cert_file: nil, client_id: '', thing_name: '',
      keep_alive: 0, enable_auto_reconnect: false)
      @host = host
      @port = port
      @ca_file = ca_file
      @key_file = key_file
      @cert_file = cert_file
      @thing_name = thing_name
      @client_id = client_id
      @keep_alive = keep_alive
      @enable_auto_reconnect = enable_auto_reconnect
      @sleep_after_subscribe = 2

      @client_for_delta = make_client(host, port, ca_file, key_file, cert_file, keep_alive)
      @client_for_update = make_client(host, port, ca_file, key_file, cert_file, keep_alive)
    end

    def connect()
      # NOTE: When @keep_alive > 0, thread calling this API may be received
      #       MQTT::ProtocolException in later.
      @client_for_delta.connect(:client_id => @client_id + 'R')
      @client_for_delta.subscribe('$aws/things/' + @thing_name + '/shadow/update/delta')
      sleep @sleep_after_subscribe
    rescue
      disconnect_without_exception
      raise
    end

    def yield()
      while true
        begin
          if block_given?
            # loop forever
            @client_for_delta.get {|topic, payload|
              delta = get_delta_from_payload(payload)
              yield(delta)
            }
          else
            # 1-shot
            topic, payload = @client_for_delta.get
            delta = get_delta_from_payload(payload)
            return delta
          end
        rescue MQTT::ProtocolException => exp
          # disconnected from server, or keep-alive timeout has occured.
          if @enable_auto_reconnect then
            # re-connect and loop again
            connect
          else
            raise AWSIotShadowDisconnectException.new("Disconnected.")
          end
        end
      end # while true
    end

    def disconnect()
      disconnect_without_exception
    end

    def update(hash)
      topic_base = '$aws/things/' + @thing_name + '/shadow/update'

      while true
        if @client_for_update.connected? then
          begin
            send_data = {:state => {:reported => hash}}.to_json
            @client_for_update.publish(topic_base, send_data)
            break # for 'while true'
          rescue
            # disconnect and loop again.
            @client_for_update.disconnect
          end
        else
          # connect and loop again.
          connect_for_update(@thing_name, @client_id)
        end
      end

      topic, payload = @client_for_update.get
      if topic != topic_base + '/accepted' then
        raise 'Rejected'
      end
    end

    def get()
      topic_base = '$aws/things/' + @thing_name + '/shadow/get'
      client_for_get = nil

      # connect
      client_for_get = make_client(@host, @port, @ca_file, @key_file, @cert_file, @keep_alive)
      client_for_get.connect(:client_id => @client_id + 'G')
      client_for_get.subscribe(topic_base + '/accepted')
      client_for_get.subscribe(topic_base + '/rejected')
      sleep @sleep_after_subscribe

      # get desired state
      client_for_get.publish(topic_base)

      topic, payload = client_for_get.get
      if topic == topic_base + '/accepted' then
        # accepted
        message = JSON.parse(payload)
        # initial state is last reported state
        state = message['state']['reported']
      else
        # rejected
        raise 'Rejected'
      end
      return state
    rescue
      raise
    ensure
      if (client_for_get && client_for_get.connected?) then
        client_for_get.disconnect
      end
    end

    def delete()
      # not yet
    end

    def reset_last_received_version()
      # not yet
    end

    def get_last_received_version()
      # not yet
    end

    def enable_discard_old_delta_msgs()
      # not yet
    end

    def disable_discard_old_delta_msgs()
      # not yet
    end

    def set_autoreconnect_status(status)
      if status then
        @enable_auto_reconnect = true
      else
        @enable_auto_reconnect = false
      end
    end

    private

    def connect_for_update(thing_name, client_id)
      @client_for_update.connect(:client_id => client_id + 'U')
      @client_for_update.subscribe('$aws/things/' + thing_name + '/shadow/update/accepted')
      @client_for_update.subscribe('$aws/things/' + thing_name + '/shadow/update/rejected')
      sleep @sleep_after_subscribe
    end

    def get_delta_from_payload(payload)
      message = JSON.parse(payload)
      delta = message['state']
      if ! delta then
        raise '"state" missing.'
      end
      return delta
    end

    def disconnect_without_exception()
      [@client_for_delta, @client_for_update].each {|c|
        begin
          if c.connected? then
            c.disconnect
          end
        rescue
          # nothing
        end
      }
    end

    def make_client(host, port, ca_file, key_file, cert_file, keep_alive)
      client = MQTT::Client.new(
        :host => host,
        :port => port,
        :version => '3.1.1',
        :keep_alive => keep_alive,
        :clean_session => true,
        :ack_timeout => 2,
        :ssl => true,
        :cert_file => cert_file,
        :key_file => key_file,
        :ca_file => ca_file
      )
      return client
    end
  end
end
