require "mqtt"

module Fluent
  class AWSIotOutput < BufferedOutput

    # First, register the plugin. NAME is the name of this plugin
    # and identifies the plugin in the configuration file.
    Plugin.register_output('awsiot', self);

    # config_param defines a parameter. You can refer a parameter via @path instance variable
    # Without :default, a parameter is required.
    # config_params :hoge, :string, :default => 'hoge'
    config_param  :certs_dir,               :string

    config_param  :root_ca,                 :string,  :default => "aws-iot-rootCA.pem",
                  :desc => "root CA"
    config_param :device_cert,              :string,  :default => "cert.pem"
    config_param :device_private_key,       :string,  :default => "privkey.pem"
    config_param :host,                     :string
    config_param :host_port,                :integer, :default => 8883
    config_param :is_clean_session,         :bool,    :default => true
    config_param :mqtt_version,             :string,  :default => "MQTT_3_1_1"
    config_param :client_id,                :string,  :default => "centuryfluent"
    config_param :is_will_msg_present,      :bool,    :default => false
    config_param :mqtt_command_timeout_ms,  :integer, :default => 2000
    config_param :tls_handshake_timeout_ms, :integer, :default => 5000
    config_param :is_ssl_hostname_verify,   :bool,    :default => true
    config_param :mqtt_qos,                 :integer, :default => 1

    def initialize
      super
    end

    # This method is called before starting.
    # 'conf' is a Hash that includes configuration parameters.
    # If the configuration is invalid, raise Fluent::ConfigError.
    def configure(conf)
      super

      # convert pathname to absolute pathname
      @root_ca = @certs_dir + "/" + @root_ca
      @device_cert = @certs_dir + "/" + @device_cert
      @device_private_key= @certs_dir + "/" + @device_private_key

      # convert mqtt_version
      case @mqtt_version
      when "MQTT_3_1_1", "3.1.1"
        @mqtt_version = "3.1.1"
      else
        @mqtt_version = "3.1.0"
      end
    end

    # This method is called when starting.
    # Open sockets or files here.
    def start
      super
      @mqttClient = makeClient
    end

    # This method is called when shutting down.
    # Shutdown the thread and close sockets or files here.
    def shutdown
      super
      if @mqttClient.connected? then
        @mqttClient.disconnect(false)
      end
    end

    # override
    # NOTE: In BufferedOutputTestDriver, this method is not called,
    #       so all data is in same chank. In Fluentd, this is work.
    def emit(tag, es, chain)
      super tag, es, chain, tag
    end

    # This method is called when an event reaches to Fluentd.
    # Convert the event to a raw string.
    # tag: string
    # time: integer
    # record: Hash
    def format(tag, time, record)
      return record.to_json + "\n"
    end

    # This method is called every flush interval. Write the buffer chunk
    # to files or databases here.
    # 'chunk' is a buffer chunk that includes multiple formatted
    # events. You can use 'data = chunk.read' to get all events and
    # 'chunk.open {|io| ... }' to get IO objects.
    #
    # NOTE: This method is called by internal thread, not Fluentd's main thread.
    #       So IO wait doesn't affect other plugins.
    # NOTE: In BufferedOutputTestDriver, chunk.key is '' (bug?).
    #       In Fluentd, chunk.key is tag (ex. 'file.etc.hosts').
    def write(chunk)
      if !( @mqttClient.connected? ) then
        log.debug "begin reconnect."
        @mqttClient.connect
        log.debug "reconnect finished."
      end

      chunk.open {|io|
        io.each{|data|
          if chunk.key == '' then
            topic = 'test.test1'
          else
            topic = chunk.key.gsub(".", "/")
          end
          ret = @mqttClient.publish(topic, data, false, @mqtt_qos)
          # When QoS=0, it returns immidiately.
          # When QoS>0, it waits ack. If failed, return -1 or raise Exception.
          if (ret == -1) then
            raise Exception("publish failed.")
          end
          sleep 0.05
        }
      }
    rescue
      if (@mqttClient.connected?) then
        @mqttClient.disconnect
      end
      raise
    end

    def makeClient()
      client = MQTT::Client.new(
        :host => @host,
        :port => @host_port,
        :version => @mqtt_version,
        :keep_alive => 0,
        :clean_session => @is_clean_session,
        :client_id=> @client_id,
        :ack_timeout => @mqtt_command_timeout_ms/1000,
        :ssl => true,
        :cert_file => @device_cert,
        :key_file => @device_private_key,
        :ca_file => @root_ca
      )
      # TODO: set ssl config(@tls_handshake_timeout_ms, @is_ssl_hostname_verify)
      return client
    end
  end
end
